# Copyright 2014 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from unittest import mock

from oslo_config import cfg

from mistral.actions import std_actions
from mistral import context as auth_context
from mistral.db.v2 import api as db_api
from mistral import exceptions as exc
from mistral.services import workbooks as wb_service
from mistral.services import workflows as wf_service
from mistral.tests.unit.engine import base
from mistral.workflow import states


# Use the set_default method to set value otherwise in certain test cases
# the change in value is not permanent.
cfg.CONF.set_default('auth_enable', False, group='pecan')


WB1 = """
---
version: '2.0'

name: wb1

workflows:
  wf1:
    type: reverse

    input:
      - param1
      - param2
    output:
      final_result: <% $.final_result %>

    tasks:
      task1:
        action: std.echo output=<% $.param1 %>
        publish:
          result1: <% task(task1).result %>

      task2:
        action: std.echo output="'<% $.param1 %> & <% $.param2 %>'"
        publish:
          final_result: <% task(task2).result %>
        requires: [task1]

  wf2:
    type: direct

    output:
      slogan: <% $.slogan %>

    tasks:
      task1:
        workflow: wf1 param1='Bonnie' param2='Clyde' task_name='task2'
        publish:
          slogan: "<% task(task1).result.final_result %> is a cool movie!"
"""

WB2 = """
---
version: '2.0'

name: wb2

workflows:
  wf1:
    type: direct

    tasks:
      task1:
        workflow: wf2

  wf2:
    type: direct

    output:
      var1: <% $.does_not_exist %>

    tasks:
        task1:
            action: std.noop
"""

WB3 = """
---
version: '2.0'

name: wb3

workflows:
  wf1:
    input:
      - wf_name
    output:
      sub_wf_out: <% $.sub_wf_out %>

    tasks:
      task1:
        workflow: <% $.wf_name %>
        publish:
          sub_wf_out: <% task(task1).result.sub_wf_out %>

  wf2:
    output:
      sub_wf_out: wf2_out

    tasks:
        task1:
            action: std.noop
"""

WB4 = """
---
version: '2.0'

name: wb4

workflows:
  wf1:
    input:
      - wf_name
      - inp
    output:
      sub_wf_out: <% $.sub_wf_out %>

    tasks:
      task1:
        workflow: <% $.wf_name %>
        input: <% $.inp %>
        publish:
          sub_wf_out: <% task(task1).result.sub_wf_out %>

  wf2:
    input:
      - inp
    output:
      sub_wf_out: <% $.inp %>

    tasks:
        task1:
            action: std.noop

"""

WB5 = """
---
version: '2.0'

name: wb5

workflows:
  wf1:
    input:
      - wf_name
      - inp
    output:
      sub_wf_out: '{{ _.sub_wf_out }}'

    tasks:
      task1:
        workflow: '{{ _.wf_name }}'
        input: '{{ _.inp }}'
        publish:
          sub_wf_out: '{{ task("task1").result.sub_wf_out }}'

  wf2:
    input:
      - inp
    output:
      sub_wf_out: '{{ _.inp }}'

    tasks:
        task1:
            action: std.noop
"""

WB6 = """
---
version: '2.0'

name: wb6

workflows:
  wf1:
    tasks:
      task1:
        workflow: wf2

  wf2:
    tasks:
      task1:
        workflow: wf3

  wf3:
    tasks:
      task1:
        action: std.noop
"""


class SubworkflowsTest(base.EngineTestCase):
    def setUp(self):
        super(SubworkflowsTest, self).setUp()

        wb_service.create_workbook_v2(WB1)
        wb_service.create_workbook_v2(WB2)
        wb_service.create_workbook_v2(WB3)
        wb_service.create_workbook_v2(WB4)
        wb_service.create_workbook_v2(WB5)
        wb_service.create_workbook_v2(WB6)

    def test_subworkflow_success(self):
        wf2_ex = self.engine.start_workflow('wb1.wf2')

        project_id = auth_context.ctx().project_id

        # Execution of 'wf2'.
        self.assertEqual(project_id, wf2_ex.project_id)
        self.assertIsNotNone(wf2_ex)
        self.assertDictEqual({}, wf2_ex.input)
        self.assertDictEqual({'namespace': '', 'env': {}}, wf2_ex.params)

        self._await(lambda: len(db_api.get_workflow_executions()) == 2, 0.5, 5)

        with db_api.transaction():
            wf_execs = db_api.get_workflow_executions()

            self.assertEqual(2, len(wf_execs))

            # Execution of 'wf2'.
            wf1_ex = self._assert_single_item(wf_execs, name='wb1.wf1')
            wf2_ex = self._assert_single_item(wf_execs, name='wb1.wf2')

            self.assertEqual(project_id, wf1_ex.project_id)
            self.assertIsNotNone(wf1_ex.task_execution_id)
            self._assert_dict_contains_subset(
                {
                    'task_name': 'task2',
                    'task_execution_id': wf1_ex.task_execution_id
                },
                wf1_ex.params
            )
            self.assertDictEqual(
                {
                    'param1': 'Bonnie',
                    'param2': 'Clyde'
                },
                wf1_ex.input
            )

        # Wait till workflow 'wf1' is completed.
        self.await_workflow_success(wf1_ex.id)

        with db_api.transaction():
            wf1_ex = db_api.get_workflow_execution(wf1_ex.id)

            wf1_output = wf1_ex.output

        self.assertDictEqual(
            {'final_result': "'Bonnie & Clyde'"},
            wf1_output
        )

        # Wait till workflow 'wf2' is completed.
        self.await_workflow_success(wf2_ex.id, timeout=4)

        with db_api.transaction():
            wf2_ex = db_api.get_workflow_execution(wf2_ex.id)

            wf2_output = wf2_ex.output

        self.assertDictEqual(
            {'slogan': "'Bonnie & Clyde' is a cool movie!"},
            wf2_output
        )

        # Check project_id in tasks.
        wf1_task_execs = db_api.get_task_executions(
            workflow_execution_id=wf1_ex.id
        )
        wf2_task_execs = db_api.get_task_executions(
            workflow_execution_id=wf2_ex.id
        )

        wf2_task1_ex = self._assert_single_item(wf1_task_execs, name='task1')
        wf1_task1_ex = self._assert_single_item(wf2_task_execs, name='task1')
        wf1_task2_ex = self._assert_single_item(wf1_task_execs, name='task2')

        self.assertEqual(project_id, wf2_task1_ex.project_id)
        self.assertEqual(project_id, wf1_task1_ex.project_id)
        self.assertEqual(project_id, wf1_task2_ex.project_id)

    @mock.patch.object(std_actions.EchoAction, 'run',
                       mock.MagicMock(side_effect=exc.ActionException))
    def test_subworkflow_error(self):
        self.engine.start_workflow('wb1.wf2')

        self._await(lambda: len(db_api.get_workflow_executions()) == 2, 0.5, 5)

        wf_execs = db_api.get_workflow_executions()

        self.assertEqual(2, len(wf_execs))

        wf1_ex = self._assert_single_item(wf_execs, name='wb1.wf1')
        wf2_ex = self._assert_single_item(wf_execs, name='wb1.wf2')

        # Wait till workflow 'wf1' is completed.
        self.await_workflow_error(wf1_ex.id)

        # Wait till workflow 'wf2' is completed, its state must be ERROR.
        self.await_workflow_error(wf2_ex.id)

    def test_subworkflow_yaql_error(self):
        wf_ex = self.engine.start_workflow('wb2.wf1')

        self.await_workflow_error(wf_ex.id)

        wf_execs = db_api.get_workflow_executions()

        self.assertEqual(2, len(wf_execs))

        wf2_ex = self._assert_single_item(wf_execs, name='wb2.wf2')

        self.assertEqual(states.ERROR, wf2_ex.state)
        self.assertIn('Can not evaluate YAQL expression', wf2_ex.state_info)

        # Ensure error message is bubbled up to the main workflow.
        wf1_ex = self._assert_single_item(wf_execs, name='wb2.wf1')

        self.assertEqual(states.ERROR, wf1_ex.state)
        self.assertIn('Can not evaluate YAQL expression', wf1_ex.state_info)

    def test_subworkflow_environment_inheritance(self):
        env = {'key1': 'abc'}

        wf2_ex = self.engine.start_workflow('wb1.wf2', env=env)

        # Execution of 'wf2'.
        self.assertIsNotNone(wf2_ex)
        self.assertDictEqual({}, wf2_ex.input)
        self.assertDictEqual(
            {'env': env, 'namespace': ''},
            wf2_ex.params
        )

        self._await(lambda: len(db_api.get_workflow_executions()) == 2, 0.5, 5)

        with db_api.transaction():
            wf_execs = db_api.get_workflow_executions()

            self.assertEqual(2, len(wf_execs))

            # Execution of 'wf1'.
            wf1_ex = self._assert_single_item(wf_execs, name='wb1.wf1')
            wf2_ex = self._assert_single_item(wf_execs, name='wb1.wf2')

            self.assertIsNotNone(wf1_ex.task_execution_id)

        # Wait till workflow 'wf1' is completed.
        self.await_workflow_success(wf1_ex.id)

        # Wait till workflow 'wf2' is completed.
        self.await_workflow_success(wf2_ex.id)

    def test_dynamic_subworkflow_wf2(self):
        ex = self.engine.start_workflow('wb3.wf1', wf_input={'wf_name': 'wf2'})

        self.await_workflow_success(ex.id)

        with db_api.transaction():
            ex = db_api.get_workflow_execution(ex.id)
            self.assertEqual({'sub_wf_out': 'wf2_out'}, ex.output)

    def test_dynamic_subworkflow_call_failure(self):
        ex = self.engine.start_workflow(
            'wb3.wf1',
            wf_input={'wf_name': 'not_existing_wf'}
        )

        self.await_workflow_error(ex.id)

        with db_api.transaction():
            ex = db_api.get_workflow_execution(ex.id)

            self.assertIn('not_existing_wf', ex.state_info)

    def test_dynamic_subworkflow_with_generic_input(self):
        self._test_dynamic_workflow_with_dict_param('wb4.wf1')

    def test_dynamic_subworkflow_with_jinja(self):
        self._test_dynamic_workflow_with_dict_param('wb5.wf1')

    def test_string_workflow_input_failure(self):
        ex = self.engine.start_workflow(
            'wb4.wf1',
            wf_input={'wf_name': 'wf2', 'inp': 'invalid_string_input'}
        )

        self.await_workflow_error(ex.id)

        with db_api.transaction():
            ex = db_api.get_workflow_execution(ex.id)

            self.assertIn('invalid_string_input', ex.state_info)

    def _test_dynamic_workflow_with_dict_param(self, wf_identifier):
        ex = self.engine.start_workflow(
            wf_identifier,
            wf_input={'wf_name': 'wf2', 'inp': {'inp': 'abc'}}
        )

        self.await_workflow_success(ex.id)

        with db_api.transaction():
            ex = db_api.get_workflow_execution(ex.id)

            self.assertEqual({'sub_wf_out': 'abc'}, ex.output)

    def test_subworkflow_root_execution_id(self):
        self.engine.start_workflow('wb6.wf1')

        self._await(lambda: len(db_api.get_workflow_executions()) == 3, 0.5, 5)

        wf_execs = db_api.get_workflow_executions()

        wf1_ex = self._assert_single_item(wf_execs, name='wb6.wf1')
        wf2_ex = self._assert_single_item(wf_execs, name='wb6.wf2')
        wf3_ex = self._assert_single_item(wf_execs, name='wb6.wf3')

        self.assertEqual(3, len(wf_execs))

        # Wait till workflow 'wf1' is completed (and all the sub-workflows
        # will be completed also).
        self.await_workflow_success(wf1_ex.id)

        with db_api.transaction():
            wf1_ex = db_api.get_workflow_execution(wf1_ex.id)
            wf2_ex = db_api.get_workflow_execution(wf2_ex.id)
            wf3_ex = db_api.get_workflow_execution(wf3_ex.id)

            self.assertIsNone(wf1_ex.root_execution_id, None)
            self.assertEqual(wf2_ex.root_execution_id, wf1_ex.id)
            self.assertEqual(wf2_ex.root_execution, wf1_ex)
            self.assertEqual(wf3_ex.root_execution_id, wf1_ex.id)
            self.assertEqual(wf3_ex.root_execution, wf1_ex)

    def test_cascade_delete(self):
        wf_text = """
        version: 2.0

        wf:
          tasks:
            task1:
              workflow: sub_wf1

            task2:
              workflow: sub_wf2

        sub_wf1:
          tasks:
            task1:
              action: std.noop

        sub_wf2:
          tasks:
            task1:
              action: std.noop
        """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf')

        self.await_workflow_success(wf_ex.id)

        self.assertEqual(3, len(db_api.get_workflow_executions()))
        self.assertEqual(4, len(db_api.get_task_executions()))
        self.assertEqual(2, len(db_api.get_action_executions()))

        # Now delete the root workflow execution and make sure that
        # all dependent objects are deleted as well.
        db_api.delete_workflow_execution(wf_ex.id)

        self.assertEqual(0, len(db_api.get_workflow_executions()))
        self.assertEqual(0, len(db_api.get_task_executions()))
        self.assertEqual(0, len(db_api.get_action_executions()))

    def test_cascade_delete_deep(self):
        wf_text = """
        version: 2.0

        wf:
          input:
            - level
          tasks:
            initial:
              action: std.noop
              on-success:
                - recurse: <% $.level > 0 %>

            recurse:
              workflow: wf
              input:
                level: <% $.level - 1 %>
        """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf', wf_input={"level": 7})

        self.await_workflow_success(wf_ex.id)

        self.assertEqual(8, len(db_api.get_workflow_executions()))

        # Now delete the root workflow execution and make sure that
        # all dependent objects are deleted as well.
        db_api.delete_workflow_execution(wf_ex.id)

        self.assertEqual(0, len(db_api.get_workflow_executions()))
        self.assertEqual(0, len(db_api.get_task_executions()))
        self.assertEqual(0, len(db_api.get_action_executions()))



# Copyright 2014 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from unittest import mock

from mistral.actions import std_actions
from mistral.db.v2 import api as db_api
from mistral.engine import engine_server
from mistral import exceptions as exc
from mistral.expressions import yaql_expression
from mistral.services import workflows as wf_service
from mistral.tests.unit.engine import base as engine_test_base


class DisabledYAQLConversionTest(engine_test_base.EngineTestCase):
    def setUp(self):
        super(DisabledYAQLConversionTest, self).setUp()

        self.override_config('auth_enable', False, 'pecan')

    def test_disabled_yaql_output_conversion(self):
        """Test YAQL expressions with disabled data conversion.

        The test is needed to make sure that if we disable YAQL data
        conversion (for both input and output), then Mistral will handle
        YAQL internal data types properly if they sneak into the Mistral
        logic as part of an expression result. Particularly, we need to
        make sure that the ORM framework (SQLAlchemy) will also be able
        to save data properly if it comes across such a type.

        NOTE:
            - set() and toSet() functions produce "frozenset" type
              internally within YAQL and it should be handled properly
              everywhere in the code including SQLAlchemy.
            - dict() produces "FrozenDict" internally but we unwrap the
              top most dict after evaluating an expression on the Mistral
              side.
        """

        # Both input and output data conversion in YAQL need to be disabled
        # so that we're sure that there won't be any surprises from YAQL
        # like some YAQL internal types included in expression results.
        self.override_config('convert_input_data', False, 'yaql')
        self.override_config('convert_output_data', False, 'yaql')

        # At this point YAQL engine has already been initialized with the
        # default value of config options. So we need to set the corresponding
        # constant to None so it gets initialized again with the new values
        # upon the first use.
        yaql_expression.YAQL_ENGINE = None

        wf_text = """---
        version: '2.0'

        wf:
          tasks:
            task1:
              publish:
                var1: <% range(0,10) %>
                var2: <% set(15) %>
                var3: <% [4, 5, 6].toSet() %>
                var4: <% {k1 => v1, k2 => v2} %>
                var5: <% dict([['a', 2], ['b', 4]]) %>
                var6: <% [1, dict(k3 => v3, k4 => v4), 3] %>
        """

        wf_service.create_workflows(wf_text)

        # Start workflow.
        wf_ex = self.engine.start_workflow('wf')

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            # Note: We need to reread execution to access related tasks.
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            tasks = wf_ex.task_executions

            t_ex = self._assert_single_item(tasks, name='task1')

            self.assertDictEqual(
                {
                    'var1': [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                    'var2': [15],
                    'var3': [4, 5, 6],
                    'var4': {'k1': 'v1', 'k2': 'v2'},
                    'var5': {'a': 2, 'b': 4},
                    'var6': [1, {'k3': 'v3', 'k4': 'v4'}, 3],
                },
                t_ex.published
            )

    def test_configuration_check(self):
        # Kill all the threads started by default and try to start an
        # instance of engine server again with the wrong configuration.
        self.kill_threads()

        self.override_config('convert_input_data', True, 'yaql')
        self.override_config('convert_output_data', False, 'yaql')

        # Setting YAQL engine to None so it reinitialized again with the
        # right values upon the next use.
        yaql_expression.YAQL_ENGINE = None

        eng_svc = engine_server.get_oslo_service(setup_profiler=False)

        self.assertRaisesWithMessage(
            exc.MistralError,
            "The config property 'yaql.convert_output_data' is set to False "
            "so 'yaql.convert_input_data' must also be set to False.",
            eng_svc.start
        )

    def test_root_context(self):
        # Both input and output data conversion in YAQL need to be disabled
        # so that we're sure that there won't be any surprises from YAQL
        # like some YAQL internal types included in expression results.
        self.override_config('convert_input_data', False, 'yaql')
        self.override_config('convert_output_data', False, 'yaql')

        # Setting YAQL engine to None so it reinitialized again with the
        # right values upon the next use.
        yaql_expression.YAQL_ENGINE = None

        wf_text = """---
        version: '2.0'

        wf:
          input:
            - param: default_val

          tasks:
            task1:
              action: std.echo output=<% $ %>
              publish:
                result: <% task().result %>
        """

        wf_service.create_workflows(wf_text)

        # Start workflow.
        wf_ex = self.engine.start_workflow('wf')

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            # Note: We need to reread execution to access related tasks.
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            t_ex = self._assert_single_item(
                wf_ex.task_executions,
                name='task1'
            )

            action_ex = t_ex.action_executions[0]

            self.assertTrue(len(action_ex.input) > 0)
            self.assertIn('output', action_ex.input)
            self.assertIn('param', action_ex.input['output'])

    def test_iterators_in_yaql_result(self):
        # Both input and output data conversion in YAQL need to be disabled
        # so that we're sure that there won't be any surprises from YAQL
        # like some YAQL internal types included in expression results.
        self.override_config('convert_input_data', False, 'yaql')
        self.override_config('convert_output_data', False, 'yaql')

        # Setting YAQL engine to None so it reinitialized again with the
        # right values upon the next use.
        yaql_expression.YAQL_ENGINE = None

        wf_text = """---
        version: '2.0'

        wf:
          input:
            - params: null

          tasks:
            task1:
              action: std.echo
              input:
                output:
                  param1:
                    <% switch($.params = null => [],
                    $.params != null =>
                    $.params.items().select({k => $[0], v => $[1]})) %>
        """

        wf_service.create_workflows(wf_text)

        wf_input = {
            'params': {
                'k1': 'v1',
                'k2': 'v2'
            }
        }

        with mock.patch.object(self.executor, 'run_action',
                               wraps=self.executor.run_action) as mocked:
            # Start workflow.
            wf_ex = self.engine.start_workflow('wf', wf_input=wf_input)

            self.await_workflow_success(wf_ex.id)

            with db_api.transaction():
                # Note: We need to reread execution to access related tasks.
                wf_ex = db_api.get_workflow_execution(wf_ex.id)

                t_ex = self._assert_single_item(
                    wf_ex.task_executions,
                    name='task1'
                )

                action_ex = t_ex.action_executions[0]

                self.assertTrue(len(action_ex.input) > 0)

            mocked.assert_called_once()

            # We need to make sure that the executor got the right action
            # input regardless of an iterator (that can only be used once)
            # present in the YAQL expression result. Let's check first 4
            # actual arguments with the executor was called, including the
            # action parameters.
            args = mocked.call_args[0]

            self.assertIsInstance(args[0], std_actions.EchoAction)
            self.assertEqual(action_ex.id, args[1])



#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import pecan
from pecan import hooks

from mistral.db.v2 import api as db_api
from mistral.services import maintenance as maintenance_service

ALLOWED_WITHOUT_AUTH = ['/', '/v2/', '/health', '/maintenance']


class MaintenanceHook(hooks.PecanHook):

    def before(self, state):
        if state.request.path in ALLOWED_WITHOUT_AUTH or \
                state.request.method == 'GET':
            return

        cluster_state = db_api.get_maintenance_status()
        is_complete_async_actions = (
            state.request.method == 'PUT' and
            '/v2/action_executions' in state.request.path and
            cluster_state == maintenance_service.PAUSING
        )

        if is_complete_async_actions or \
                cluster_state == maintenance_service.RUNNING:
            return

        msg = "Current Mistral state is {}. Method is not allowed".format(
            cluster_state
        )

        pecan.abort(
            status_code=423,
            detail=msg,
            headers={'Server-Error-Message': msg}
        )



#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistral.db.v2 import api as db_api
from mistral.services import workflows as wf_service
from mistral.tests.unit.engine import base
from mistral.workflow import states
from oslo_config import cfg

# Use the set_default method to set value otherwise in certain test cases
# the change in value is not permanent.
cfg.CONF.set_default('auth_enable', False, group='pecan')


class DirectWorkflowEngineTest(base.EngineTestCase):

    def _run_workflow(self, wf_text, wf_input=None,
                      expected_state=states.SUCCESS):
        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf', wf_input=wf_input)

        self.await_workflow_state(wf_ex.id, expected_state)

        return db_api.get_workflow_execution(wf_ex.id)

    def test_replace(self):
        self.override_config('merge_strategy', 'replace',
                             group='engine')
        wf_text = """
        version: '2.0'
        wf:
          input:
            - aa:
                bb: wf_ex_input
                cc: wf_ex_input
                zz: wf_ex_input
          output:
            aa: <% $.aa %>
          tasks:
            task1:
              action: std.echo
              # emulate some action result
              input:
                output:
                  cc: task1_res
                  dd: task1_res
              on-success: [task2]
              publish:
                aa:
                  cc: <% task().result["cc"] %>
                  dd: <% task().result["dd"] %>
            task2:
              action: std.echo
              # emulate some action result
              input:
                output:
                  bb: task2_res
              publish:
                aa:
                  bb: <% task().result["bb"] %>
        """

        wf_ex = self._run_workflow(wf_text)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)
            wf_ex_res = wf_ex.output

            task1 = self._assert_single_item(wf_ex.task_executions,
                                             name='task1')
            task1_publish = task1.published

        expected_wf_ex_res = {
            "aa": {
                "bb": "task2_res"
            }
        }
        self.assertDictEqual(expected_wf_ex_res, wf_ex_res)

        expected_task1_res = {
            "aa": {
                "cc": "task1_res",
                "dd": "task1_res"
            }
        }
        self.assertDictEqual(expected_task1_res, task1_publish)

    def test_merge_input_and_publish(self):
        self.override_config('merge_strategy', 'merge',
                             group='engine')

        wf_text = """
        version: '2.0'

        wf:
          input:
            - aa:
                bb:
                  bb_1: wf_ex_input
                  bb_2: wf_ex_input
                cc: wf_ex_input
          output:
            aa: <% $.aa %>
          tasks:
            task1:
              action: std.echo
              # emulate some action result
              input:
                output:
                  cc: task_res
                  dd: task_res
                  bb_2: task_res
              publish:
                aa:
                  bb:
                    bb_2: <% task().result["bb_2"] %>
                  cc: <% task().result["cc"] %>
                  dd: <% task().result["dd"] %>
        """

        wf_ex = self._run_workflow(wf_text)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)
            wf_ex_res = wf_ex.output

        expected_wf_ex_res = {
            "aa": {
                "bb": {
                    "bb_1": "wf_ex_input",
                    "bb_2": "task_res"
                },
                "cc": "task_res",
                "dd": "task_res"
            }
        }
        self.assertDictEqual(expected_wf_ex_res, wf_ex_res)

    def test_merge_publish_two_tasks(self):
        self.override_config('merge_strategy', 'merge',
                             group='engine')
        wf_text = """
        version: '2.0'
        wf:
          input:
            - aa:
                bb: wf_ex_input
                cc: wf_ex_input
                zz: wf_ex_input
          output:
            aa: <% $.aa %>
          tasks:
            task1:
              action: std.echo
              # emulate some action result
              input:
                output:
                  cc: task1_res
                  dd: task1_res
              on-success: [task2]
              publish:
                aa:
                  cc: <% task().result["cc"] %>
                  dd: <% task().result["dd"] %>
            task2:
              action: std.echo
              # emulate some action result
              input:
                output:
                  bb: task2_res
              publish:
                aa:
                  bb: <% task().result["bb"] %>
        """

        wf_ex = self._run_workflow(wf_text)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)
            wf_ex_res = wf_ex.output

        expected_wf_ex_res = {
            "aa": {
                "bb": "task2_res",
                "cc": "task1_res",
                "dd": "task1_res",
                "zz": "wf_ex_input"
            }
        }
        self.assertDictEqual(expected_wf_ex_res, wf_ex_res)

    def test_merge_publish_result_from_join(self):
        self.override_config('merge_strategy', 'merge',
                             group='engine')
        wf_text = """
        version: '2.0'
        wf:
          input:
            - aa:
                aa: wf_ex_input
                bb: wf_ex_input
                cc: wf_ex_input
                zz: wf_ex_input
                dd: wf_ex_input
          output:
            aa: <% $.aa %>
          tasks:
            task0:
              action: std.echo
              # emulate some action result
              input:
                output:
                  res: task0_res
              on-success: [task1, task2]
              publish:
                aa:
                  aa: <% task().result["res"] %>
                  aa_1: <% task().result["res"] %>
            task1:
              action: std.echo
              # emulate some action result
              input:
                output:
                  res: task1_res
              on-success: [task3]
              publish:
                aa:
                  bb: <% task().result["res"] %>
                  bb_1: <% task().result["res"] %>
            task2:
              action: std.echo
              input:
                output:
                  res: task2_res
              on-success: [task3]
              publish:
                aa:
                  cc: <% task().result["res"] %>
                  cc_1: <% task().result["res"] %>
            task3:
              action: std.echo
              join: all
              input:
                output:
                  res: task3_res
              publish:
                aa:
                  zz: <% $.aa.bb_1 + $.aa.cc_1 + $.aa.aa_1 %>
                  zz_1: <% task().result["res"] %>
        """

        wf_ex = self._run_workflow(wf_text)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)
            wf_ex_res = wf_ex.output

        expected_wf_ex_res = {
            "aa": {
                "aa": "task0_res",
                "aa_1": "task0_res",
                "bb": "task1_res",
                "bb_1": "task1_res",
                "cc": "task2_res",
                "cc_1": "task2_res",
                "zz": "task1_restask2_restask0_res",
                "zz_1": "task3_res",
                "dd": "wf_ex_input"
            }
        }
        self.assertDictEqual(expected_wf_ex_res, wf_ex_res)

    def test_merge_publish_two_parallel_tasks(self):
        self.override_config('merge_strategy', 'merge',
                             group='engine')
        wf_text = """
        version: '2.0'
        wf:
          input:
            - aa:
                bb: wf_ex_input
                cc: wf_ex_input
                dd: wf_ex_input
                ee: wf_ex_input
                zz: wf_ex_input
          output:
            aa: <% $.aa %>
          tasks:
            task_1_1:
              action: std.echo
              # emulate some action result
              input:
                output:
                  res: task_1_1_res
              on-success: [task_2_1]
              publish:
                aa:
                  bb: <% task().result["res"] %>
            task_1_2:
              action: std.echo
              # emulate some action result
              input:
                output:
                  res: task_1_2_res
              on-success: [task_2_1]
              publish:
                aa:
                  cc: <% task().result["res"] %>
            task_2_1:
              action: std.echo
              # emulate some action result
              join: all
              input:
                output:
                  res: task_2_1_res
              on-success: [task_3_1]
              publish:
                aa:
                  dd: <% task().result["res"] %>
            task_3_1:
              action: std.echo
              # emulate some action result
              input:
                output:
                  res: task_3_1_res
              publish:
                aa:
                  ee: <% task().result["res"] %>
        """

        wf_ex = self._run_workflow(wf_text)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)
            wf_ex_res = wf_ex.output

        expected_wf_ex_res = {
            "aa": {
                "bb": "task_1_1_res",
                "cc": "task_1_2_res",
                "dd": "task_2_1_res",
                "ee": "task_3_1_res",
                "zz": "wf_ex_input"
            }
        }
        self.assertDictEqual(expected_wf_ex_res, wf_ex_res)



# Copyright 2014 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import copy
from unittest import mock

import sqlalchemy as sa

from mistral.actions import adhoc
from mistral.db.v2 import api as db_api
from mistral.db.v2.sqlalchemy import models
from mistral.lang import parser as spec_parser
from mistral.services import adhoc_actions
from mistral.tests.unit.api import base

from mistral_lib.actions.providers import composite
from mistral_lib import utils


ADHOC_ACTION_YAML_A = """
---
version: '2.0'

a_action:
  description: My super cool action.
  tags: ['test', 'v2']
  base: std.echo
  base-input:
    output: "{$.str1}{$.str2}"
  input:
    - str1
    - str2
"""

ADHOC_ACTION_YAML_B = """
---
version: '2.0'

b_action:
  description: My super cool action.
  tags: ['test', 'v2']
  base: std.echo
  base-input:
    output: "{$.str1}{$.str2}"
  input:
    - str1
    - str2
"""

ADHOC_ACTION_YAML = """
---
version: '2.0'

my_action:
  description: My super cool action.
  tags: ['test', 'v2']
  base: std.echo
  base-input:
    output: "{$.str1}{$.str2}"
  input:
    - str1
    - str2
"""

ACTION_DEFINITION_INVALID_NO_BASE = """
---
version: '2.0'

my_action:
  description: My super cool action.
  tags: ['test', 'v2']

  base-input:
    output: "{$.str1}{$.str2}"
"""

ACTION_DEFINITION_INVALID_YAQL = """
---
version: '2.0'

my_action:
  description: My super cool action.
  tags: ['test', 'v2']
  base: std.echo
  base-input:
    output: <% $. %>
"""

ACTION_DSL_PARSE_EXCEPTION = """
---
%
"""

ACTION_SPEC = spec_parser.get_action_list_spec_from_yaml(ADHOC_ACTION_YAML)[0]

ACTION_DEF_VALUES = {
    'id': '123e4567-e89b-12d3-a456-426655440000',
    'name': 'my_action',
    'is_system': False,
    'description': 'My super cool action.',
    'tags': ['test', 'v2'],
    'definition': ADHOC_ACTION_YAML,
    'spec': ACTION_SPEC.to_dict(),
    'input': '',
    'project_id': None,
    'scope': 'public',
    'namespace': None
}


ACTION_DEF = models.ActionDefinition()
ACTION_DEF.update(ACTION_DEF_VALUES)

ACTION_DESC = adhoc.AdHocActionDescriptor(ACTION_DEF)


UPDATED_ADHOC_ACTION_YAML = """
---
version: '2.0'

my_action:
  description: My super cool action.
  base: std.echo
  base-input:
    output: "{$.str1}{$.str2}{$.str3}"
"""

UPDATED_ACTION_DEF = copy.copy(ACTION_DEF)
UPDATED_ACTION_DEF['definition'] = UPDATED_ADHOC_ACTION_YAML
UPDATED_ACTION = copy.deepcopy(ACTION_DEF_VALUES)
UPDATED_ACTION['definition'] = UPDATED_ADHOC_ACTION_YAML

MOCK_ACTIONS = mock.MagicMock(return_value=[ACTION_DEF])
MOCK_UPDATED_ACTION = mock.MagicMock(return_value=UPDATED_ACTION_DEF)


class TestActionsController(base.APITest):
    def check_adhoc_action_json(self, action_json):
        self.assertIsNotNone(action_json)
        self.assertIsInstance(action_json, dict)

        action_name = action_json['name']

        action_def = db_api.get_action_definition(action_name)

        self.assertIsNotNone(
            action_def,
            'Ad-hoc action definition does not exist [name=%s]' % action_name
        )

        # Compare action JSON with the state of the corresponding
        # persistent object.
        for k, v in action_json.items():
            self.assertEqual(v, utils.datetime_to_str(getattr(action_def, k)))

    def test_get(self):
        # Create an adhoc action for the purpose of the test.
        adhoc_actions.create_actions(ADHOC_ACTION_YAML)

        resp = self.app.get('/v2/actions/my_action')

        self.assertEqual(200, resp.status_int)

        self.check_adhoc_action_json(resp.json)

    @mock.patch.object(db_api, 'load_action_definition')
    def test_get_operational_error(self, mocked_get):
        # Create an adhoc action for the purpose of the test.
        adhoc_actions.create_actions(ADHOC_ACTION_YAML)

        action_def = db_api.get_action_definition('my_action')

        mocked_get.side_effect = [
            # Emulating DB OperationalError
            sa.exc.OperationalError('Mock', 'mock', 'mock'),
            action_def  # Successful run
        ]

        resp = self.app.get('/v2/actions/my_action')

        self.assertEqual(200, resp.status_int)

        self.check_adhoc_action_json(resp.json)

    def test_get_not_found(self):
        # This time we don't create an action in DB upfront.
        resp = self.app.get('/v2/actions/my_action', expect_errors=True)

        self.assertEqual(404, resp.status_int)

    def test_get_by_id(self):
        # NOTE(rakhmerov): We can't support this case anymore because
        # action descriptors can now be identified only by names and
        # namespaces.
        pass

    def test_get_within_project_id(self):
        # Create an adhoc action for the purpose of the test.
        adhoc_actions.create_actions(ADHOC_ACTION_YAML)

        # We should not be able to change 'project_id' even with a
        # direct DB call.
        db_api.update_action_definition(
            'my_action',
            {'project_id': 'foobar'}
        )

        resp = self.app.get('/v2/actions/my_action')

        self.assertEqual(200, resp.status_int)
        self.assertEqual('<default-project>', resp.json['project_id'])

    def test_put(self):
        # Create an adhoc action for the purpose of the test.
        adhoc_actions.create_actions(ADHOC_ACTION_YAML)

        resp = self.app.put(
            '/v2/actions',
            UPDATED_ADHOC_ACTION_YAML,
            headers={'Content-Type': 'text/plain'}
        )

        self.assertEqual(200, resp.status_int)

        self.check_adhoc_action_json(resp.json['actions'][0])

    def test_put_public(self):
        # Create an adhoc action for the purpose of the test.
        adhoc_actions.create_actions(ADHOC_ACTION_YAML)

        resp = self.app.put(
            '/v2/actions?scope=public',
            UPDATED_ADHOC_ACTION_YAML,
            headers={'Content-Type': 'text/plain'}
        )

        action_json = resp.json['actions'][0]

        self.assertEqual(200, resp.status_int)
        self.assertEqual('public', action_json['scope'])

        self.check_adhoc_action_json(action_json)

    def test_put_not_found(self):
        resp = self.app.put(
            '/v2/actions',
            UPDATED_ADHOC_ACTION_YAML,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(404, resp.status_int)

    def test_put_system(self):
        # Create an adhoc action for the purpose of the test.
        adhoc_actions.create_actions(ADHOC_ACTION_YAML)

        db_api.update_action_definition('my_action', {'is_system': True})

        resp = self.app.put(
            '/v2/actions',
            ADHOC_ACTION_YAML,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)
        self.assertIn(
            'Attempt to modify a system action: my_action',
            resp.body.decode()
        )

    def test_post(self):
        resp = self.app.post(
            '/v2/actions',
            ADHOC_ACTION_YAML,
            headers={'Content-Type': 'text/plain'}
        )

        self.assertEqual(201, resp.status_int)

        self.check_adhoc_action_json(resp.json['actions'][0])

    def test_post_public(self):
        resp = self.app.post(
            '/v2/actions?scope=public',
            ADHOC_ACTION_YAML,
            headers={'Content-Type': 'text/plain'}
        )

        self.assertEqual(201, resp.status_int)

        self.check_adhoc_action_json(resp.json['actions'][0])
        self.assertEqual('public', resp.json['actions'][0]['scope'])

    def test_post_wrong_scope(self):
        resp = self.app.post(
            '/v2/actions?scope=unique',
            ADHOC_ACTION_YAML,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)
        self.assertIn("Scope must be one of the following", resp.body.decode())

    def test_post_dup(self):
        resp = self.app.post(
            '/v2/actions',
            ADHOC_ACTION_YAML,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(201, resp.status_int)

        # Try to create it again.
        resp = self.app.post(
            '/v2/actions',
            ADHOC_ACTION_YAML,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(409, resp.status_int)

    def test_delete(self):
        # Create an adhoc action for the purpose of the test.
        adhoc_actions.create_actions(ADHOC_ACTION_YAML)

        self.assertIsNotNone(db_api.load_action_definition('my_action'))

        resp = self.app.delete('/v2/actions/my_action')

        self.assertEqual(204, resp.status_int)

        self.assertIsNone(db_api.load_action_definition('my_action'))

    def test_delete_not_found(self):
        resp = self.app.delete('/v2/actions/my_action', expect_errors=True)

        self.assertEqual(404, resp.status_int)

    def test_get_all(self):
        # Create an adhoc action for the purpose of the test.
        adhoc_actions.create_actions(ADHOC_ACTION_YAML_B)
        adhoc_actions.create_actions(ADHOC_ACTION_YAML_A)
        adhoc_actions.create_actions(ADHOC_ACTION_YAML)

        resp = self.app.get('/v2/actions')

        self.assertEqual(200, resp.status_int)

        actions_json = resp.json['actions']

        # Verify they are sorted alphabetically
        self.assertEqual(actions_json[0]['name'], 'a_action')
        self.assertEqual(actions_json[1]['name'], 'b_action')
        self.assertEqual(actions_json[2]['name'], 'my_action')

        # There will be 'std.' actions and the one we've just created.
        self.assertGreater(len(actions_json), 1)

        # Let's check some of the well-known 'std.' actions.
        self._assert_single_item(actions_json, name='std.echo')
        self._assert_single_item(actions_json, name='std.ssh')
        self._assert_single_item(actions_json, name='std.fail')
        self._assert_single_item(actions_json, name='std.noop')
        self._assert_single_item(actions_json, name='std.async_noop')

        # Now let's check the ad-hoc action data.
        adhoc_action_json = self._assert_single_item(
            actions_json,
            name='my_action'
        )

        self.check_adhoc_action_json(adhoc_action_json)

    @mock.patch.object(db_api, 'get_action_definitions')
    def test_get_all_operational_error(self, mocked_get_all):
        # Create an adhoc action for the purpose of the test.
        adhoc_actions.create_actions(ADHOC_ACTION_YAML)

        action_def = db_api.get_action_definition('my_action')

        mocked_get_all.side_effect = [
            # Emulating DB OperationalError
            sa.exc.OperationalError('Mock', 'mock', 'mock'),
            [action_def]  # Successful run
        ]

        resp = self.app.get('/v2/actions')

        actions_json = resp.json['actions']

        # There will be 'std.' actions and the one we've just created.
        self.assertGreater(len(actions_json), 1)

        # Let's check some of the well-known 'std.' actions.
        self._assert_single_item(actions_json, name='std.echo')
        self._assert_single_item(actions_json, name='std.ssh')
        self._assert_single_item(actions_json, name='std.fail')
        self._assert_single_item(actions_json, name='std.noop')
        self._assert_single_item(actions_json, name='std.async_noop')

        # Now let's check the ad-hoc action data.
        adhoc_action_json = self._assert_single_item(
            actions_json,
            name='my_action'
        )

        self.check_adhoc_action_json(adhoc_action_json)

    @mock.patch.object(
        composite.CompositeActionProvider,
        'find_all',
        mock.MagicMock(return_value=[])
    )
    def test_get_all_empty(self):
        resp = self.app.get('/v2/actions')

        self.assertEqual(200, resp.status_int)

        self.assertEqual(0, len(resp.json['actions']))

    def test_get_all_filtered(self):
        # First check that w/o filters the result set
        # will contain more than 1 item.
        resp = self.app.get('/v2/actions')

        self.assertEqual(200, resp.status_int)
        self.assertGreater(len(resp.json['actions']), 1)

        # Now we'll filter it by action name.
        resp = self.app.get('/v2/actions?name=std.echo')

        self.assertEqual(200, resp.status_int)
        self.assertEqual(1, len(resp.json['actions']))

    def test_get_all_pagination(self):
        # Create an adhoc action for the purpose of the test.
        adhoc_actions.create_actions(ADHOC_ACTION_YAML)

        resp = self.app.get(
            '/v2/actions?limit=1&sort_keys=created_at&sort_dirs=desc')

        self.assertEqual(200, resp.status_int)
        self.assertIn('next', resp.json)
        self.assertEqual(1, len(resp.json['actions']))

        self.check_adhoc_action_json(resp.json['actions'][0])

        param_dict = utils.get_dict_from_string(
            resp.json['next'].split('?')[1],
            delimiter='&'
        )

        action_def = db_api.get_action_definition('my_action')

        # TODO(rakhmerov): In this case we can't use IDs for marker because
        # in general we don't identify action descriptors with IDs.
        expected_dict = {
            'marker': action_def.id,
            'limit': 1,
            'sort_keys': 'created_at,id',
            'sort_dirs': 'desc,asc'
        }

        self.assertTrue(
            set(expected_dict.items()).issubset(set(param_dict.items()))
        )

    def test_get_all_sort_date_asc(self):
        adhoc_actions.create_actions(ADHOC_ACTION_YAML)
        resp = self.app.get('/v2/actions?sort_keys=created_at&sort_dirs=asc')
        self.assertEqual(200, resp.status_int)

    def test_get_all_sort_date_desc(self):
        adhoc_actions.create_actions(ADHOC_ACTION_YAML_A)
        adhoc_actions.create_actions(ADHOC_ACTION_YAML_B)
        resp = self.app.get('/v2/actions?sort_keys=created_at&sort_dirs=desc')
        self.assertEqual(200, resp.status_int)

    def test_get_all_pagination_marker(self):
        adhoc_actions.create_actions(ADHOC_ACTION_YAML_B)
        adhoc_actions.create_actions(ADHOC_ACTION_YAML_A)
        adhoc_actions.create_actions(ADHOC_ACTION_YAML)

        resp = self.app.get('/v2/actions?limit=1')

        self.assertEqual(200, resp.status_int)
        self.assertEqual(1, len(resp.json['actions']))
        self.assertEqual(resp.json['actions'][0]['name'], 'a_action')

        resp = self.app.get('/v2/actions?marker=my_action&limit=2')

        self.assertEqual(200, resp.status_int)
        self.assertEqual(2, len(resp.json['actions']))
        self.assertEqual(resp.json['actions'][0]['name'], 'my_action')
        self.assertEqual(resp.json['actions'][1]['name'], 'std.async_noop')

    def test_get_all_pagination_limit_negative(self):
        resp = self.app.get(
            '/v2/actions?limit=-1&sort_keys=id,name&sort_dirs=asc,asc',
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)

        self.assertIn("Limit must be positive", resp.body.decode())

    def test_get_all_pagination_limit_not_integer(self):
        resp = self.app.get(
            '/v2/actions?limit=1.1&sort_keys=id,name&sort_dirs=asc,asc',
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)

        self.assertIn("unable to convert to int", resp.body.decode())

    def test_get_all_pagination_invalid_sort_dirs_length(self):
        resp = self.app.get(
            '/v2/actions?limit=1&sort_keys=id,name&sort_dirs=asc,asc,asc',
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)

        self.assertIn(
            "Length of sort_keys must be equal or greater than sort_dirs",
            resp.body.decode()
        )

    def test_get_all_pagination_unknown_direction(self):
        resp = self.app.get(
            '/v2/actions?limit=1&sort_keys=id&sort_dirs=nonexist',
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)

        self.assertIn("Unknown sort direction", resp.body.decode())

    def test_validate(self):
        resp = self.app.post(
            '/v2/actions/validate',
            ADHOC_ACTION_YAML,
            headers={'Content-Type': 'text/plain'}
        )

        self.assertEqual(200, resp.status_int)
        self.assertTrue(resp.json['valid'])

    def test_validate_invalid_model_exception(self):
        resp = self.app.post(
            '/v2/actions/validate',
            ACTION_DEFINITION_INVALID_NO_BASE,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(200, resp.status_int)
        self.assertFalse(resp.json['valid'])
        self.assertIn("Invalid DSL", resp.json['error'])

    def test_validate_dsl_parse_exception(self):
        resp = self.app.post(
            '/v2/actions/validate',
            ACTION_DSL_PARSE_EXCEPTION,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(200, resp.status_int)
        self.assertFalse(resp.json['valid'])
        self.assertIn("Definition could not be parsed", resp.json['error'])

    def test_validate_yaql_parse_exception(self):
        resp = self.app.post(
            '/v2/actions/validate',
            ACTION_DEFINITION_INVALID_YAQL,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(200, resp.status_int)
        self.assertFalse(resp.json['valid'])
        self.assertIn("unexpected end of statement",
                      resp.json['error'])

    def test_validate_empty(self):
        resp = self.app.post(
            '/v2/actions/validate',
            '',
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(200, resp.status_int)
        self.assertFalse(resp.json['valid'])
        self.assertIn("Invalid DSL", resp.json['error'])



# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


import pecan
from pecan import rest

from mistral import exceptions as exc


class SpecValidationController(rest.RestController):

    def __init__(self, parser):
        super(SpecValidationController, self).__init__()

        self._parse_func = parser

    @pecan.expose('json')
    def post(self):
        """Validate a spec."""
        definition = pecan.request.text

        try:
            self._parse_func(definition, validate=True)
        except exc.DSLParsingException as e:
            return {'valid': False, 'error': str(e)}

        return {'valid': True}



# Copyright 2019 - Nokia Networks.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from datetime import datetime
from oslo_log import log as logging
from pecan import rest
import wsmeext.pecan as wsme_pecan

from mistral.api.controllers.v2 import resources
from mistral.api.controllers.v2 import types
from mistral.db.v2 import api as db_api
from mistral.db.v2.sqlalchemy import models as db_models
from mistral.utils import rest_utils
from mistral.workflow import states


LOG = logging.getLogger(__name__)

ESTIMATED_TIME_QUERY_LIMIT = 20


def create_workflow_execution_entry(wf_ex):
    return resources.WorkflowExecutionReportEntry.from_db_model(wf_ex)


def create_task_execution_entry(task_ex):
    return resources.TaskExecutionReportEntry.from_db_model(task_ex)


def create_action_execution_entry(action_ex):
    return resources.ActionExecutionReportEntry.from_db_model(action_ex)


def update_statistics_with_task(stat, task_ex):
    if task_ex.state == states.RUNNING:
        stat.increment_running()
    elif task_ex.state == states.SUCCESS:
        stat.increment_success()
    elif task_ex.state == states.ERROR:
        stat.increment_error()
    elif task_ex.state == states.IDLE:
        stat.increment_idle()
    elif task_ex.state == states.PAUSED:
        stat.increment_paused()


def analyse_task_execution(task_ex_id, stat, filters, cur_depth):
    with db_api.transaction():
        task_ex = db_api.get_task_execution(task_ex_id)

        if filters['errors_only'] and task_ex.state != states.ERROR:
            return None

        update_statistics_with_task(stat, task_ex)

        entry = create_task_execution_entry(task_ex)

        child_executions = task_ex.executions

    if 'retry_task_policy' in task_ex.runtime_context:
        retry_ctx = task_ex.runtime_context['retry_task_policy']

        entry.retry_count = retry_ctx['retry_no']

    entry.action_executions = []
    entry.workflow_executions = []

    for c_ex in child_executions:
        if isinstance(c_ex, db_models.ActionExecution):
            entry.action_executions.append(
                create_action_execution_entry(c_ex)
            )
        else:
            entry.workflow_executions.append(
                analyse_workflow_execution(c_ex.id, stat, filters, cur_depth)
            )

    return entry


def analyse_workflow_execution(wf_ex_id, stat, filters, cur_depth):
    with db_api.transaction():
        wf_ex = db_api.get_workflow_execution(wf_ex_id)

        entry = create_workflow_execution_entry(wf_ex)

        max_depth = filters['max_depth']

        # Don't get deeper into the workflow task executions if
        # maximum depth is defined and the current depth exceeds it.
        if 0 <= max_depth < cur_depth:
            return entry

        task_execs = wf_ex.task_executions

    entry.task_executions = []

    for t_ex in task_execs:
        task_exec_entry = analyse_task_execution(
            t_ex.id,
            stat,
            filters,
            cur_depth + 1
        )

        if task_exec_entry:
            entry.task_executions.append(task_exec_entry)

    return entry


def analyse_task_statistics_only(task_ex_id, stat, filters, cur_depth):
    with db_api.transaction():
        task_ex = db_api.get_task_execution(task_ex_id)

        if filters['errors_only'] and task_ex.state != states.ERROR:
            return

        update_statistics_with_task(stat, task_ex)

        child_executions = task_ex.executions

    for c_ex in child_executions:
        if not isinstance(c_ex, db_models.ActionExecution):
            analyse_execution_statistics_only(
                c_ex.id,
                stat,
                filters,
                cur_depth
            )


def analyse_execution_statistics_only(wf_ex_id, stat, filters, cur_depth):

    max_depth = filters['max_depth']

    if 0 <= max_depth < cur_depth:
        return

    with db_api.transaction():
        wf_ex = db_api.get_workflow_execution(wf_ex_id)

        task_execs = wf_ex.task_executions

    for t_ex in task_execs:
        analyse_task_statistics_only(
            t_ex.id,
            stat,
            filters,
            cur_depth + 1
        )


def calculate_estimated_left_time_for_exec(wf_ex, prev_wf_exs):
    if len(prev_wf_exs) == 0:
        return -1

    ex_runtime_sec = (datetime.now() - wf_ex.created_at).seconds
    average_runtime_sec = sum((ex.updated_at - ex.created_at).seconds
                              for ex in prev_wf_exs) / len(prev_wf_exs)
    estimated_time_sec = average_runtime_sec - ex_runtime_sec

    if estimated_time_sec < 0:
        return 1

    return estimated_time_sec


def estimate_time_for_execution(ex_id):
    with db_api.transaction():
        wf_ex = db_api.get_workflow_execution(ex_id)

        if wf_ex.state == states.RUNNING:
            prev_wf_exs = db_api.get_workflow_executions(
                limit=ESTIMATED_TIME_QUERY_LIMIT,
                workflow_id=wf_ex.workflow_id,
                state=states.SUCCESS
            )
            return calculate_estimated_left_time_for_exec(wf_ex, prev_wf_exs)

    return 0


def build_report(wf_ex_id, filters):
    report = resources.ExecutionReport()

    stat = resources.ExecutionReportStatistics()
    stat.estimated_time = estimate_time_for_execution(wf_ex_id)

    report.statistics = stat

    if not filters['statistics_only']:
        report.root_workflow_execution = analyse_workflow_execution(
            wf_ex_id,
            stat,
            filters,
            0
        )
    else:
        analyse_execution_statistics_only(
            wf_ex_id,
            stat,
            filters,
            0
        )

    return report


class ExecutionReportController(rest.RestController):
    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(resources.ExecutionReport, types.uuid, bool, int,
                         bool)
    def get(self, workflow_execution_id, errors_only=False, max_depth=-1,
            statistics_only=False):
        """Return workflow execution report.

        :param workflow_execution_id: The ID of the workflow execution to
            generate a report for.
        :param errors_only: Optional. If True, only error paths of the
            execution tree are included into the report. The root execution
            (with the specified id) is always included, but its tasks may
            or may not be included depending on this flag's value.
        :param max_depth: Optional. Limits the depth of recursion while
            obtaining the execution tree. That is, subworkflows of what
            maximum depth will be included into the report. If a value of the
            flag is a negative number then no limit is set.
            The root execution has depth 0 so if the flag is 0 then only
            the root execution, its tasks and their actions will be included.
            If some of the tasks in turn run workflows then these subworkflows
            will be also included but without their tasks. The algorithm will
            fully analyse their tasks only if max_depth is greater than zero.
        :param statistics_only: Optional. If True, only the statistics will be
            returned.
        """

        LOG.info(
            "Fetch execution report [workflow_execution_id=%s]",
            workflow_execution_id
        )

        filters = {
            'errors_only': errors_only,
            'max_depth': max_depth,
            'statistics_only': statistics_only
        }

        return build_report(workflow_execution_id, filters)



# Copyright 2015 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
# Copyright 2016 - Nokia Networks.
# Copyright 2016 - Brocade Communications Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_log import log as logging
from osprofiler import profiler
import traceback as tb

from mistral.db import utils as db_utils
from mistral.db.v2 import api as db_api
from mistral.db.v2.sqlalchemy import models
from mistral.engine import post_tx_queue
from mistral.engine import tasks
from mistral.engine import workflow_handler as wf_handler
from mistral import exceptions as exc
from mistral.lang import parser as spec_parser
from mistral.scheduler import base as sched_base
from mistral.workflow import base as wf_base
from mistral.workflow import commands as wf_cmds
from mistral.workflow import states

# TODO(rakhmerov): At some point we need to completely switch to
# exceptions from mistral_lib. For the time being though we'll
# have to catch MistralException from both 'mistral' and 'mistral-lib'.
from mistral_lib import exceptions as mistral_lib_exc


"""Responsible for running tasks and handling results."""

LOG = logging.getLogger(__name__)

_REFRESH_TASK_STATE_PATH = (
    'mistral.engine.task_handler._refresh_task_state'
)

_SCHEDULED_ON_ACTION_COMPLETE_PATH = (
    'mistral.engine.task_handler._scheduled_on_action_complete'
)

_SCHEDULED_ON_ACTION_UPDATE_PATH = (
    'mistral.engine.task_handler._scheduled_on_action_update'
)


@profiler.trace('task-handler-run-task', hide_args=True)
def run_task(task_ex_id, waiting, triggered_by, rerun, reset, first_run=False):
    """Runs workflow task.

    :param task_ex_id: Task Execution id
    :param waiting: Task waiting param
    :param triggered_by:
    :param rerun:
    :param reset:
    :param first_run:
    """

    task_ex = db_api.get_task_execution(task_ex_id)
    wf_spec = spec_parser.get_workflow_spec_by_execution_id(
        task_ex.workflow_execution_id
    )
    task = _build_task_after_rpc(wf_spec, task_ex, waiting, triggered_by,
                                 rerun, reset)

    try:
        if task.waiting and task.rerun:
            task.set_state(states.WAITING, 'Task is waiting.')

            _schedule_refresh_task_state(task.task_ex.id)

        task.run(first_run)
    except (exc.MistralException, mistral_lib_exc.MistralException) as e:
        wf_ex = task_ex.workflow_execution

        msg = (
            "Failed to run task [error=%s, wf=%s, task=%s]:\n%s" %
            (e, wf_ex.name, task_ex.name, tb.format_exc())
        )

        force_fail_task(task_ex, msg, task=task)
        return

    _check_affected_tasks(task)


@profiler.trace('task-handler-skip-task', hide_args=True)
def skip_task(wf_cmd):
    """Skip workflow task.

    :param wf_cmd: Workflow command.
    """
    task = _build_task_from_command(wf_cmd)
    task.complete(states.SKIPPED, "Task was skipped.", skip=True)
    _check_affected_tasks(task)
    return


@profiler.trace('task-handler-create-task', hide_args=True)
def create_task(wf_cmd, first_run):
    """Creates workflow task.

    :param wf_cmd: Workflow command.
    """

    task = _build_task_from_command(wf_cmd)

    if task.waiting and task.rerun:
        task.set_state(states.WAITING, 'Task is waiting.')

        _schedule_refresh_task_state(task.task_ex.id)

    if first_run:
        task.create_new()

    return task


def mark_task_running(task_ex, wf_spec):
    task = build_task_from_execution(wf_spec, task_ex)

    task.set_state(states.RUNNING, None, False)


@profiler.trace('task-handler-on-action-complete', hide_args=True)
def _on_action_complete(action_ex):
    """Handles action completion event.

    :param action_ex: Action execution.
    """

    task_ex = action_ex.task_execution

    if not task_ex:
        return

    task_spec = spec_parser.get_task_spec(task_ex.spec)

    wf_ex = task_ex.workflow_execution

    task = _create_task(
        wf_ex,
        spec_parser.get_workflow_spec_by_execution_id(wf_ex.id),
        task_spec,
        task_ex.in_context,
        task_ex
    )

    try:
        task.on_action_complete(action_ex)
    except (exc.MistralException, mistral_lib_exc.MistralException) as e:
        wf_ex = task_ex.workflow_execution

        msg = ("Failed to handle action completion [error=%s, wf=%s, task=%s,"
               " action=%s]:\n%s" %
               (e, wf_ex.name, task_ex.name, action_ex.name, tb.format_exc()))

        force_fail_task(task_ex, msg, task=task)

        return

    _check_affected_tasks(task)


@profiler.trace('task-handler-on-action-update', hide_args=True)
def _on_action_update(action_ex):
    """Handles action update event.

    :param action_ex: Action execution.
    """

    task_ex = action_ex.task_execution

    if not task_ex:
        return

    task_spec = spec_parser.get_task_spec(task_ex.spec)

    wf_ex = task_ex.workflow_execution

    task = _create_task(
        wf_ex,
        spec_parser.get_workflow_spec_by_execution_id(wf_ex.id),
        task_spec,
        task_ex.in_context,
        task_ex
    )

    try:
        task.on_action_update(action_ex)

        if states.is_paused(action_ex.state):
            wf_handler.pause_workflow(wf_ex)

        if states.is_running(action_ex.state):
            # If any subworkflow of the parent workflow is paused,
            # then keep the parent workflow execution paused.
            for task_ex in wf_ex.task_executions:
                if states.is_paused(task_ex.state):
                    return

            # Otherwise if no other subworkflow is paused,
            # then resume the parent workflow execution.
            wf_handler.resume_workflow(wf_ex)

    except (exc.MistralException, mistral_lib_exc.MistralException) as e:
        wf_ex = task_ex.workflow_execution

        msg = ("Failed to handle action update [error=%s, wf=%s, task=%s,"
               " action=%s]:\n%s" %
               (e, wf_ex.name, task_ex.name, action_ex.name, tb.format_exc()))

        force_fail_task(task_ex, msg, task=task)

        return

    _check_affected_tasks(task)


def force_fail_task(task_ex, msg, task=None):
    """Forces the given task to fail.

    This method implements the 'forced' task fail without giving a chance
    to a workflow controller to handle the error. Its main purpose is to
    reflect errors caused by workflow structure (errors 'publish', 'on-xxx'
    clauses etc.) rather than failed actions. If such an error happens
    we should also force the entire workflow to fail. I.e., this kind of
    error must be propagated to a higher level, to the workflow.

    :param task_ex: Task execution.
    :param msg: Error message.
    :param task: Task object. Optional.
    """

    LOG.error(msg)

    if not task:
        wf_spec = spec_parser.get_workflow_spec_by_execution_id(
            task_ex.workflow_execution_id
        )

        task = build_task_from_execution(wf_spec, task_ex)

    task.set_state(states.ERROR, msg)

    wf_handler.force_fail_workflow(task_ex.workflow_execution, msg)


def continue_task(task_ex):
    if not task_ex:
        return

    wf_spec = spec_parser.get_workflow_spec_by_execution_id(
        task_ex.workflow_execution_id
    )

    task = build_task_from_execution(wf_spec, task_ex)

    try:
        with db_api.named_lock('continue-task-%s' % task_ex.id):
            task.set_state(states.RUNNING, None)
            task.run()
    except exc.MistralException as e:
        wf_ex = task_ex.workflow_execution

        msg = (
            "Failed to run task [error=%s, wf=%s, task=%s]:\n%s" %
            (e, wf_ex.name, task_ex.name, tb.format_exc())
        )

        force_fail_task(task_ex, msg, task=task)

        return

    _check_affected_tasks(task)


def complete_task(task_ex, state, state_info):
    if not task_ex:
        return

    wf_spec = spec_parser.get_workflow_spec_by_execution_id(
        task_ex.workflow_execution_id
    )

    task = build_task_from_execution(wf_spec, task_ex)

    try:
        task.complete(state, state_info)
    except (exc.MistralException, mistral_lib_exc.MistralException) as e:
        wf_ex = task_ex.workflow_execution

        msg = (
            "Failed to complete task [error=%s, wf=%s, task=%s]:\n%s" %
            (e, wf_ex.name, task_ex.name, tb.format_exc())
        )

        force_fail_task(task_ex, msg, task=task)

        return

    _check_affected_tasks(task)


@profiler.trace('task-handler-check-affected-tasks', hide_args=True)
def _check_affected_tasks(task):
    # TODO(rakhmerov): this method should eventually move into
    # the class Task. The obvious signal is the only argument
    # that it takes.
    if not task.is_completed():
        return

    task_ex = task.task_ex

    wf_ex = task_ex.workflow_execution

    if states.is_completed(wf_ex.state):
        return

    wf_spec = spec_parser.get_workflow_spec_by_execution_id(
        task_ex.workflow_execution_id
    )

    wf_ctrl = wf_base.get_controller(wf_ex, wf_spec)

    affected_task_execs = wf_ctrl.find_indirectly_affected_task_executions(
        task_ex.name
    )

    def _schedule_if_needed(t_ex_id):
        # NOTE(rakhmerov): we need to minimize the number of scheduled jobs
        # that refresh state of "join" tasks. We'll check if corresponding
        # jobs are already scheduled. Note that we must ignore scheduled jobs
        # that are currently being processed because of a possible race with
        # the transaction that deletes scheduled jobs, i.e. the job may still
        # exist in DB (the deleting transaction didn't commit yet) but it has
        # already been processed and the task state hasn't changed.
        sched = sched_base.get_system_scheduler()

        jobs_exist = sched.has_scheduled_jobs(
            key=_get_refresh_state_job_key(t_ex_id),
            processing=False
        )

        if not jobs_exist:
            _schedule_refresh_task_state(t_ex_id)

    for t_ex in affected_task_execs:
        post_tx_queue.register_operation(
            _schedule_if_needed,
            args=[t_ex.id],
            in_tx=True
        )


def build_task_from_execution(wf_spec, task_ex):
    return _create_task(
        task_ex.workflow_execution,
        wf_spec,
        wf_spec.get_task(task_ex.name),
        task_ex.in_context,
        task_ex
    )


def _build_task_after_rpc(wf_spec, task_ex, waiting, triggered_by, rerun,
                          reset):
    task = _create_task(
        task_ex.workflow_execution,
        wf_spec,
        wf_spec.get_task(task_ex.name),
        task_ex.in_context,
        task_ex,
        waiting=waiting == states.WAITING,
        triggered_by=triggered_by,
        rerun=rerun
    )

    if reset:
        task.reset()

    return task


@profiler.trace('task-handler-build-task-from-command', hide_args=True)
def _build_task_from_command(cmd):
    if isinstance(cmd, wf_cmds.RunExistingTask):
        task = _create_task(
            cmd.wf_ex,
            cmd.wf_spec,
            spec_parser.get_task_spec(cmd.task_ex.spec),
            cmd.ctx,
            task_ex=cmd.task_ex,
            unique_key=cmd.task_ex.unique_key,
            waiting=cmd.task_ex.state == states.WAITING,
            triggered_by=cmd.triggered_by,
            rerun=cmd.rerun
        )

        if cmd.reset:
            task.reset()

        return task

    if isinstance(cmd, wf_cmds.RunTask):
        task = _create_task(
            cmd.wf_ex,
            cmd.wf_spec,
            cmd.task_spec,
            cmd.ctx,
            unique_key=cmd.unique_key,
            waiting=cmd.is_waiting(),
            triggered_by=cmd.triggered_by
        )

        return task

    if isinstance(cmd, wf_cmds.SkipTask):
        task = _create_task(
            cmd.wf_ex,
            cmd.wf_spec,
            spec_parser.get_task_spec(cmd.task_ex.spec),
            cmd.ctx,
            task_ex=cmd.task_ex,
            unique_key=cmd.task_ex.unique_key,
            triggered_by=cmd.triggered_by,
        )

        return task

    raise exc.MistralError('Unsupported workflow command: %s' % cmd)


def _create_task(wf_ex, wf_spec, task_spec, ctx, task_ex=None,
                 unique_key=None, waiting=False, triggered_by=None,
                 rerun=False):
    if task_spec.get_with_items():
        cls = tasks.WithItemsTask
    else:
        cls = tasks.RegularTask

    return cls(
        wf_ex,
        wf_spec,
        task_spec,
        ctx,
        task_ex=task_ex,
        unique_key=unique_key,
        waiting=waiting,
        triggered_by=triggered_by,
        rerun=rerun
    )


@db_utils.retry_on_db_error
@post_tx_queue.run
@profiler.trace('task-handler-refresh-task-state', hide_args=True)
def _refresh_task_state(task_ex_id):
    with db_api.transaction():
        task_ex = db_api.load_task_execution(task_ex_id)

        if not task_ex:
            return

        if (states.is_completed(task_ex.state)
                or task_ex.state == states.RUNNING):
            return

        wf_ex = task_ex.workflow_execution

        if states.is_completed(wf_ex.state):
            return

        wf_spec = spec_parser.get_workflow_spec_by_execution_id(
            task_ex.workflow_execution_id
        )

        wf_ctrl = wf_base.get_controller(wf_ex, wf_spec)

        with db_api.named_lock(task_ex.id):
            # NOTE: we have to use this lock to prevent two (or more) such
            # methods from changing task state and starting its action or
            # workflow. Checking task state outside of this section is a
            # performance optimization because locking is pretty expensive.
            db_api.refresh(task_ex)

            if (states.is_completed(task_ex.state)
                    or task_ex.state == states.RUNNING):
                return

            log_state = wf_ctrl.get_logical_task_state(task_ex)

            state = log_state.state
            state_info = log_state.state_info

            # Update 'triggered_by' because it could have changed.
            task_ex.runtime_context['triggered_by'] = log_state.triggered_by

            if state == states.RUNNING:
                continue_task(task_ex)
            elif state == states.ERROR:
                complete_task(task_ex, state, state_info)
            elif state == states.WAITING:
                LOG.info(
                    "Task execution is still in WAITING state"
                    " [task_ex_id=%s, task_name=%s]",
                    task_ex_id,
                    task_ex.name
                )
            else:
                # Must never get here.
                raise RuntimeError(
                    'Unexpected logical task state [task_ex_id=%s, '
                    'task_name=%s, state=%s]' %
                    (task_ex_id, task_ex.name, state)
                )


def _schedule_refresh_task_state(task_ex_id, delay=0):
    """Schedules task preconditions check.

    This method provides transactional decoupling of task preconditions
    check from events that can potentially satisfy those preconditions.

    It's needed in non-locking model in order to avoid 'phantom read'
    phenomena when reading state of multiple tasks to see if a task that
    depends on them can start. Just starting a separate transaction
    without using scheduler is not safe due to concurrency window that
    we'll have in this case (time between transactions) whereas scheduler
    is a special component that is designed to be resistant to failures.

    :param task_ex_id: Task execution ID.
    :param delay: Delay.
    """

    sched = sched_base.get_system_scheduler()

    job = sched_base.SchedulerJob(
        run_after=delay,
        func_name=_REFRESH_TASK_STATE_PATH,
        func_args={'task_ex_id': task_ex_id},
        key=_get_refresh_state_job_key(task_ex_id)
    )

    sched.schedule(job)


def _get_refresh_state_job_key(task_ex_id):
    return 'th_r_t_s-%s' % task_ex_id


@db_utils.retry_on_db_error
@post_tx_queue.run
def _scheduled_on_action_complete(action_ex_id, wf_action):
    with db_api.transaction():
        if wf_action:
            action_ex = db_api.load_workflow_execution(action_ex_id)
        else:
            action_ex = db_api.load_action_execution(action_ex_id)

        if action_ex:
            _on_action_complete(action_ex)


def schedule_on_action_complete(action_ex, delay=0):
    """Schedules task completion check.

    This method provides transactional decoupling of action completion from
    task completion check. It's needed in non-locking model in order to
    avoid 'phantom read' phenomena when reading state of multiple actions
    to see if a task is completed. Just starting a separate transaction
    without using scheduler is not safe due to concurrency window that we'll
    have in this case (time between transactions) whereas scheduler is a
    special component that is designed to be resistant to failures.

    :param action_ex: Action execution.
    :param delay: Minimum amount of time before task completion check
        should be made.
    """

    # Optimization to avoid opening a new transaction if it's not needed.
    if not action_ex.task_execution.spec.get('with-items'):
        _on_action_complete(action_ex)

        return

    sched = sched_base.get_system_scheduler()

    job = sched_base.SchedulerJob(
        run_after=delay,
        func_name=_SCHEDULED_ON_ACTION_COMPLETE_PATH,
        func_args={
            'action_ex_id': action_ex.id,
            'wf_action': isinstance(action_ex, models.WorkflowExecution)
        },
        key='th_on_a_c-%s' % action_ex.task_execution_id
    )

    sched.schedule(job)


@db_utils.retry_on_db_error
@post_tx_queue.run
def _scheduled_on_action_update(action_ex_id, wf_action):
    with db_api.transaction():
        if wf_action:
            action_ex = db_api.load_workflow_execution(action_ex_id)
        else:
            action_ex = db_api.load_action_execution(action_ex_id)

        if action_ex:
            _on_action_update(action_ex)


def schedule_on_action_update(action_ex, delay=0):
    """Schedules task update check.

    This method provides transactional decoupling of action update from
    task update check. It's needed in non-locking model in order to
    avoid 'phantom read' phenomena when reading state of multiple actions
    to see if a task is updated. Just starting a separate transaction
    without using scheduler is not safe due to concurrency window that we'll
    have in this case (time between transactions) whereas scheduler is a
    special component that is designed to be resistant to failures.

    :param action_ex: Action execution.
    :param delay: Minimum amount of time before task update check
        should be made.
    """

    # Optimization to avoid opening a new transaction if it's not needed.
    if not action_ex.task_execution.spec.get('with-items'):
        _on_action_update(action_ex)

        return

    sched = sched_base.get_system_scheduler()

    job = sched_base.SchedulerJob(
        run_after=delay,
        func_name=_SCHEDULED_ON_ACTION_UPDATE_PATH,
        func_args={
            'action_ex_id': action_ex.id,
            'wf_action': isinstance(action_ex, models.WorkflowExecution)
        },
        key='th_on_a_u-%s' % action_ex.task_execution_id
    )

    sched.schedule(job)



# Copyright 2018 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Increase delayed_calls_v2.auth_context column size from JsonDictType to
JsonMediumDictType

Revision ID: 030
Revises: 029
Create Date: 2018-08-07 08:35:57.609328

"""

# revision identifiers, used by Alembic.
revision = '030'
down_revision = '029'

from alembic import op
from mistral.db.sqlalchemy import types as st


def upgrade():
    # Changing column type from JsonDictType to JsonLongDictType
    op.alter_column('delayed_calls_v2', 'auth_context',
                    type_=st.JsonMediumDictType())



# Copyright 2016 - Brocade Communications Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import abc

from oslo_config import cfg
from stevedore import driver

from mistral import exceptions as exc


_IMPL_AUTH_HANDLER = None


def get_auth_handler():
    auth_type = cfg.CONF.auth_type

    global _IMPL_AUTH_HANDLER

    if not _IMPL_AUTH_HANDLER:
        mgr = driver.DriverManager(
            'mistral.auth',
            auth_type,
            invoke_on_load=True
        )

        _IMPL_AUTH_HANDLER = mgr.driver

    return _IMPL_AUTH_HANDLER


class AuthHandler(object, metaclass=abc.ABCMeta):
    """Abstract base class for an authentication plugin."""

    @abc.abstractmethod
    def authenticate(self, req):
        raise exc.UnauthorizedException()



# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from unittest import mock

from oslo_config import cfg
import requests

from mistral.actions import std_actions
from mistral.db.v2 import api as db_api
from mistral.services import workflows as wf_service
from mistral.tests.unit import base as test_base
from mistral.tests.unit.engine import base
from mistral.workflow import states


# Use the set_default method to set value otherwise in certain test cases
# the change in value is not permanent.
cfg.CONF.set_default('auth_enable', False, group='pecan')

ENV = {
    '__actions': {
        'std.http': {
            'auth': 'librarian:password123',
            'timeout': 30,
        }
    }
}

EXPECTED_ENV_AUTH = ('librarian', 'password123')

WORKFLOW1 = """
---
version: "2.0"

wf1:
  tasks:
    task1:
      action: std.http url="https://api.library.org/books"
      publish:
        result: <% $ %>
"""

WORKFLOW2 = """
---
version: "2.0"

wf2:
  tasks:
    task1:
      action: std.http url="https://api.library.org/books" timeout=60
      publish:
        result: <% $ %>
"""

WORKFLOW1_WITH_ITEMS = """
---
version: "2.0"

wf1_with_items:
  input:
    - links

  tasks:
    task1:
      with-items: link in <% $.links %>
      action: std.http url=<% $.link %>
      publish:
        result: <% $ %>
"""

WORKFLOW2_WITH_ITEMS = """
---
version: "2.0"

wf2_with_items:
  input:
    - links

  tasks:
    task1:
      with-items: link in <% $.links %>
      action: std.http url=<% $.link %> timeout=60
      publish:
        result: <% $ %>
"""


class ActionDefaultTest(base.EngineTestCase):
    @mock.patch.object(
        requests, 'request',
        mock.MagicMock(return_value=test_base.FakeHTTPResponse('', 200, 'OK')))
    @mock.patch.object(
        std_actions.HTTPAction, 'is_sync',
        mock.MagicMock(return_value=True))
    def test_action_defaults_from_env(self):
        wf_service.create_workflows(WORKFLOW1)

        wf_ex = self.engine.start_workflow('wf1', env=ENV)

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            self.assertEqual(states.SUCCESS, wf_ex.state)
            self._assert_single_item(wf_ex.task_executions, name='task1')

        requests.request.assert_called_with(
            'GET',
            'https://api.library.org/books',
            params=None,
            data=None,
            json=None,
            headers=None,
            cookies=None,
            allow_redirects=None,
            proxies=None,
            verify=None,
            auth=EXPECTED_ENV_AUTH,
            timeout=ENV['__actions']['std.http']['timeout']
        )

    @mock.patch.object(
        requests, 'request',
        mock.MagicMock(return_value=test_base.FakeHTTPResponse('', 200, 'OK')))
    @mock.patch.object(
        std_actions.HTTPAction, 'is_sync',
        mock.MagicMock(return_value=True))
    def test_action_defaults_from_env_not_applied(self):
        wf_service.create_workflows(WORKFLOW2)

        wf_ex = self.engine.start_workflow('wf2', env=ENV)

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            self.assertEqual(states.SUCCESS, wf_ex.state)
            self._assert_single_item(wf_ex.task_executions, name='task1')

        requests.request.assert_called_with(
            'GET', 'https://api.library.org/books',
            params=None, data=None, json=None, headers=None, cookies=None,
            allow_redirects=None, proxies=None, verify=None,
            auth=EXPECTED_ENV_AUTH,
            timeout=60
        )

    @mock.patch.object(
        requests, 'request',
        mock.MagicMock(return_value=test_base.FakeHTTPResponse('', 200, 'OK')))
    @mock.patch.object(
        std_actions.HTTPAction, 'is_sync',
        mock.MagicMock(return_value=True))
    def test_with_items_action_defaults_from_env(self):
        wf_service.create_workflows(WORKFLOW1_WITH_ITEMS)

        wf_input = {
            'links': [
                'https://api.library.org/books',
                'https://api.library.org/authors'
            ]
        }

        wf_ex = self.engine.start_workflow(
            'wf1_with_items',
            wf_input=wf_input,
            env=ENV
        )

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            self.assertEqual(states.SUCCESS, wf_ex.state)
            self._assert_single_item(wf_ex.task_executions, name='task1')

            calls = [mock.call('GET', url, params=None, data=None,
                               json=None, headers=None, cookies=None,
                               allow_redirects=None, proxies=None,
                               auth=EXPECTED_ENV_AUTH, verify=None,
                               timeout=ENV['__actions']['std.http']['timeout'])
                     for url in wf_input['links']]

        requests.request.assert_has_calls(calls, any_order=True)

    @mock.patch.object(
        requests, 'request',
        mock.MagicMock(return_value=test_base.FakeHTTPResponse('', 200, 'OK')))
    @mock.patch.object(
        std_actions.HTTPAction, 'is_sync',
        mock.MagicMock(return_value=True))
    def test_with_items_action_defaults_from_env_not_applied(self):
        wf_service.create_workflows(WORKFLOW2_WITH_ITEMS)

        wf_input = {
            'links': [
                'https://api.library.org/books',
                'https://api.library.org/authors'
            ]
        }

        wf_ex = self.engine.start_workflow(
            'wf2_with_items',
            wf_input=wf_input,
            env=ENV
        )

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            self.assertEqual(states.SUCCESS, wf_ex.state)
            self._assert_single_item(wf_ex.task_executions, name='task1')

        calls = [mock.call('GET', url, params=None, data=None,
                           json=None, headers=None, cookies=None,
                           allow_redirects=None, proxies=None,
                           auth=EXPECTED_ENV_AUTH, verify=None,
                           timeout=60)
                 for url in wf_input['links']]

        requests.request.assert_has_calls(calls, any_order=True)



# Copyright 2015 Huawei Technologies Co., Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_config import cfg
from oslo_log import log as logging
from pecan import rest
import tooz.coordination
import wsmeext.pecan as wsme_pecan

from mistral.api import access_control as acl
from mistral.api.controllers.v2 import resources
from mistral import context
from mistral import exceptions as exc
from mistral.service import coordination
from mistral.utils import rest_utils


LOG = logging.getLogger(__name__)


class ServicesController(rest.RestController):
    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(resources.Services)
    def get_all(self):
        """Return all services."""
        acl.enforce('services:list', context.ctx())

        LOG.debug("Fetch services.")

        if not cfg.CONF.coordination.backend_url:
            raise exc.CoordinationNotSupportedException("Service API "
                                                        "is not supported.")

        service_coordinator = coordination.get_service_coordinator()

        if not service_coordinator.is_active():
            raise exc.CoordinationException(
                "Failed to connect to coordination backend."
            )

        # Should be the same as LAUNCH_OPTIONS in launch.py
        # At the moment there is a duplication, need to solve it.
        # We cannot depend on launch.py since it uses eventlet monkey patch
        # under wsgi it causes problems
        mistral_services = {'api', 'engine', 'executor',
                            'event-engine', 'notifier'}
        services_list = []
        service_group = ['%s_group' % i for i in mistral_services]

        try:
            for group in service_group:
                members = service_coordinator.get_members(group)

                members_list = [
                    resources.Service.from_dict(
                        {
                            'type': group,
                            'name': member
                        }
                    )
                    for member in members
                ]

                services_list.extend(members_list)
        except tooz.coordination.ToozError as e:
            # In the scenario of network interruption or manually shutdown
            # connection shutdown, ToozError will be raised.
            raise exc.CoordinationException(
                "Failed to get service members from coordination backend. %s"
                % str(e)
            )

        return resources.Services(services=services_list)



# Copyright (c) 2016 Intel Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import futurist
from mistral import context
from mistral import exceptions as exc
from mistral.tests.unit.rpc.kombu import base
from mistral.tests.unit.rpc.kombu import fake_kombu
from unittest import mock

import socket
from stevedore import driver

with mock.patch.dict('sys.modules', kombu=fake_kombu):
    from mistral.rpc.kombu import kombu_server


class TestException(exc.MistralError):
    pass


class KombuServerTest(base.KombuTestCase):

    def setUp(self):
        super(KombuServerTest, self).setUp()

        self.conf = mock.MagicMock()
        self.server = kombu_server.KombuRPCServer(self.conf)
        self.ctx = type('context', (object,), {'to_dict': lambda self: {}})()

    def test_is_running_is_running(self):
        self.server._running.set()
        self.assertTrue(self.server.is_running)

    def test_is_running_is_not_running(self):
        self.server._running.clear()
        self.assertFalse(self.server.is_running)

    def test_stop(self):
        self.server.stop()
        self.assertFalse(self.server.is_running)

    def test_publish_message(self):
        body = 'body'
        reply_to = 'reply_to'
        corr_id = 'corr_id'
        type = 'type'

        acquire_mock = mock.MagicMock()
        fake_kombu.producer.acquire.return_value = acquire_mock

        enter_mock = mock.MagicMock()
        acquire_mock.__enter__.return_value = enter_mock

        self.server.publish_message(body, reply_to, corr_id, type)
        enter_mock.publish.assert_called_once_with(
            body={'body': '"body"'},
            exchange='openstack',
            routing_key=reply_to,
            correlation_id=corr_id,
            type=type,
            serializer='json'
        )

    def test_run_launch_successfully(self):
        acquire_mock = mock.MagicMock()
        acquire_mock.drain_events.side_effect = TestException()
        fake_kombu.connection.acquire.return_value = acquire_mock

        self.assertRaises(TestException, self.server._run, 'blocking')
        self.assertTrue(self.server.is_running)

    def test_run_launch_successfully_than_stop(self):

        def side_effect(*args, **kwargs):
            self.assertTrue(self.server.is_running)
            raise KeyboardInterrupt

        acquire_mock = mock.MagicMock()
        acquire_mock.drain_events.side_effect = side_effect
        fake_kombu.connection.acquire.return_value = acquire_mock

        self.server._run('blocking')
        self.assertFalse(self.server.is_running)
        self.assertEqual(self.server._sleep_time, 1)

    def test_run_socket_error_reconnect(self):

        def side_effect(*args, **kwargs):
            if acquire_mock.drain_events.call_count == 1:
                raise socket.error()
            raise TestException()

        acquire_mock = mock.MagicMock()
        acquire_mock.drain_events.side_effect = side_effect
        fake_kombu.connection.acquire.return_value = acquire_mock

        self.assertRaises(TestException, self.server._run, 'blocking')
        self.assertEqual(self.server._sleep_time, 1)

    def test_run_socket_timeout_still_running(self):

        def side_effect(*args, **kwargs):
            if acquire_mock.drain_events.call_count == 0:
                raise socket.timeout()
            raise TestException()

        acquire_mock = mock.MagicMock()
        acquire_mock.drain_events.side_effect = side_effect
        fake_kombu.connection.acquire.return_value = acquire_mock

        self.assertRaises(
            TestException,
            self.server._run,
            'blocking'
        )
        self.assertTrue(self.server.is_running)

    def test_run_keyboard_interrupt_not_running(self):
        acquire_mock = mock.MagicMock()
        acquire_mock.drain_events.side_effect = KeyboardInterrupt()
        fake_kombu.connection.acquire.return_value = acquire_mock

        self.assertIsNone(self.server.run())
        self.assertFalse(self.server.is_running)

    @mock.patch.object(
        kombu_server.KombuRPCServer,
        '_on_message',
        mock.MagicMock()
    )
    @mock.patch.object(kombu_server.KombuRPCServer, 'publish_message')
    def test__on_message_safe_message_processing_ok(self, publish_message):
        message = mock.MagicMock()

        self.server._on_message_safe(None, message)

        self.assertEqual(message.ack.call_count, 1)
        self.assertEqual(publish_message.call_count, 0)

    @mock.patch.object(kombu_server.KombuRPCServer, '_on_message')
    @mock.patch.object(kombu_server.KombuRPCServer, 'publish_message')
    def test__on_message_safe_message_processing_raise(
            self,
            publish_message,
            _on_message
    ):
        reply_to = 'reply_to'
        correlation_id = 'corr_id'
        message = mock.MagicMock()
        message.properties = {
            'reply_to': reply_to,
            'correlation_id': correlation_id
        }

        test_exception = TestException()
        _on_message.side_effect = test_exception

        self.server._on_message_safe(None, message)

        self.assertEqual(message.ack.call_count, 1)
        self.assertEqual(publish_message.call_count, 1)

    @mock.patch.object(
        kombu_server.KombuRPCServer,
        '_get_rpc_method',
        mock.MagicMock(return_value=None)
    )
    def test__on_message_rpc_method_not_found(self):
        request = {
            'rpc_ctx': {},
            'rpc_method': 'not_found_method',
            'arguments': {}
        }

        message = mock.MagicMock()
        message.properties = {
            'reply_to': None,
            'correlation_id': None
        }

        self.assertRaises(
            exc.MistralException,
            self.server._on_message,
            request,
            message
        )

    @mock.patch.object(kombu_server.KombuRPCServer, 'publish_message')
    @mock.patch.object(kombu_server.KombuRPCServer, '_get_rpc_method')
    @mock.patch('mistral.context.MistralContext.from_dict')
    def test__on_message_is_async(self, mock_get_context, get_rpc_method,
                                  publish_message):
        result = 'result'
        request = {
            'async': True,
            'rpc_ctx': {},
            'rpc_method': 'found_method',
            'arguments': self.server._serialize_message({
                'a': 1,
                'b': 2
            })
        }

        message = mock.MagicMock()
        message.properties = {
            'reply_to': None,
            'correlation_id': None
        }
        message.delivery_info.get.return_value = False

        rpc_method = mock.MagicMock(return_value=result)
        get_rpc_method.return_value = rpc_method

        ctx = context.MistralContext()
        mock_get_context.return_value = ctx

        self.server._on_message(request, message)

        rpc_method.assert_called_once_with(
            rpc_ctx=ctx,
            a=1,
            b=2
        )
        self.assertEqual(publish_message.call_count, 0)

    @mock.patch.object(kombu_server.KombuRPCServer, 'publish_message')
    @mock.patch.object(kombu_server.KombuRPCServer, '_get_rpc_method')
    @mock.patch('mistral.context.MistralContext.from_dict')
    def test__on_message_is_sync(self, mock_get_context, get_rpc_method,
                                 publish_message):
        result = 'result'
        request = {
            'async': False,
            'rpc_ctx': {},
            'rpc_method': 'found_method',
            'arguments': self.server._serialize_message({
                'a': 1,
                'b': 2
            })
        }

        reply_to = 'reply_to'
        correlation_id = 'corr_id'
        message = mock.MagicMock()
        message.properties = {
            'reply_to': reply_to,
            'correlation_id': correlation_id
        }
        message.delivery_info.get.return_value = False

        rpc_method = mock.MagicMock(return_value=result)
        get_rpc_method.return_value = rpc_method

        ctx = context.MistralContext()
        mock_get_context.return_value = ctx

        self.server._on_message(request, message)

        rpc_method.assert_called_once_with(
            rpc_ctx=ctx,
            a=1,
            b=2
        )
        publish_message.assert_called_once_with(
            result,
            reply_to,
            correlation_id
        )

    def test__prepare_worker(self):
        self.server._prepare_worker('blocking')
        self.assertEqual(
            futurist.SynchronousExecutor,
            type(self.server._worker)
        )

        self.server._prepare_worker('threading')
        self.assertEqual(
            futurist.ThreadPoolExecutor,
            type(self.server._worker)
        )

        self.server._prepare_worker('eventlet')
        self.assertEqual(
            futurist.GreenThreadPoolExecutor,
            type(self.server._worker)
        )

    @mock.patch('stevedore.driver.DriverManager')
    def test__prepare_worker_no_valid_executor(self, driver_manager_mock):

        driver_manager_mock.side_effect = driver.NoMatches()

        self.assertRaises(
            driver.NoMatches,
            self.server._prepare_worker,
            'non_valid_executor'
        )



# Copyright 2015 Huawei Technologies Co., Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from unittest import mock

from oslo_config import cfg

from mistral.service import coordination
from mistral.tests.unit import base


class ServiceCoordinatorTest(base.BaseTest):
    def test_start(self):
        cfg.CONF.set_default(
            'backend_url',
            'zake://',
            'coordination'
        )

        coordinator = coordination.ServiceCoordinator('fake_id')
        coordinator.start()

        self.assertTrue(coordinator.is_active())

    def test_start_without_backend(self):
        cfg.CONF.set_default('backend_url', None, 'coordination')

        coordinator = coordination.ServiceCoordinator()
        coordinator.start()

        self.assertFalse(coordinator.is_active())

    def test_stop_not_active(self):
        cfg.CONF.set_default('backend_url', None, 'coordination')

        coordinator = coordination.ServiceCoordinator()
        coordinator.start()

        coordinator.stop()

        self.assertFalse(coordinator.is_active())

    def test_stop(self):
        cfg.CONF.set_default(
            'backend_url',
            'zake://',
            'coordination'
        )

        coordinator = coordination.ServiceCoordinator()
        coordinator.start()

        coordinator.stop()

        self.assertFalse(coordinator.is_active())

    def test_join_group_not_active(self):
        cfg.CONF.set_default('backend_url', None, 'coordination')

        coordinator = coordination.ServiceCoordinator()
        coordinator.start()

        coordinator.join_group('fake_group')
        members = coordinator.get_members('fake_group')

        self.assertFalse(coordinator.is_active())

        self.assertEqual(0, len(members))

    def test_join_group_and_get_members(self):
        cfg.CONF.set_default(
            'backend_url',
            'zake://',
            'coordination'
        )

        coordinator = coordination.ServiceCoordinator(my_id='fake_id')
        coordinator.start()

        coordinator.join_group('fake_group')
        members = coordinator.get_members('fake_group')

        self.assertEqual(1, len(members))
        self.assertCountEqual(('fake_id'.encode("latin-1"),), members)

    def test_join_group_and_leave_group(self):
        cfg.CONF.set_default(
            'backend_url',
            'zake://',
            'coordination'
        )

        coordinator = coordination.ServiceCoordinator(my_id='fake_id')
        coordinator.start()

        coordinator.join_group('fake_group')
        members_before = coordinator.get_members('fake_group')

        coordinator.leave_group('fake_group')
        members_after = coordinator.get_members('fake_group')

        self.assertEqual(1, len(members_before))
        self.assertEqual(set(['fake_id'.encode("latin-1")]), members_before)

        self.assertEqual(0, len(members_after))
        self.assertEqual(set([]), members_after)


class ServiceTest(base.BaseTest):
    def setUp(self):
        super(ServiceTest, self).setUp()

        # Re-initialize the global service coordinator object, in order to use
        # new coordination configuration.
        coordination.cleanup_service_coordinator()

    @mock.patch('mistral_lib.utils.get_process_identifier',
                return_value='fake_id')
    def test_register_membership(self, mock_get_identifier):
        cfg.CONF.set_default('backend_url', 'zake://', 'coordination')

        srv = coordination.Service('fake_group')
        srv.register_membership()

        self.addCleanup(srv.stop)

        srv_coordinator = coordination.get_service_coordinator()

        self.assertIsNotNone(srv_coordinator)
        self.assertTrue(srv_coordinator.is_active())

        members = srv_coordinator.get_members('fake_group')

        mock_get_identifier.assert_called_once_with()
        self.assertEqual(set(['fake_id'.encode("latin-1")]), members)



# Copyright 2015 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
# Copyright 2020 Nokia Software.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import contextlib

from oslo_config import cfg
from oslo_db import api as db_api

_BACKEND_MAPPING = {
    'sqlalchemy': 'mistral.db.v2.sqlalchemy.api',
}

IMPL = db_api.DBAPI('sqlalchemy', backend_mapping=_BACKEND_MAPPING)

CONF = cfg.CONF


def setup_db():
    IMPL.setup_db()


def drop_db():
    IMPL.drop_db()


# Transaction control.


def start_tx():
    IMPL.start_tx()


def commit_tx():
    IMPL.commit_tx()


def rollback_tx():
    IMPL.rollback_tx()


def end_tx():
    IMPL.end_tx()


@contextlib.contextmanager
def transaction(read_only=False):
    with IMPL.transaction(read_only):
        yield


def refresh(model):
    IMPL.refresh(model)


def expire_all():
    IMPL.expire_all()


# Locking.


def acquire_lock(model, id):
    return IMPL.acquire_lock(model, id)


# Workbooks.

def get_workbook(name, namespace, fields=()):
    return IMPL.get_workbook(name, namespace=namespace, fields=fields)


def load_workbook(name, namespace, fields=()):
    """Unlike get_workbook this method is allowed to return None."""
    return IMPL.load_workbook(name, namespace=namespace, fields=fields)


def get_workbooks(limit=None, marker=None, sort_keys=None,
                  sort_dirs=None, fields=None, **kwargs):
    return IMPL.get_workbooks(
        limit=limit,
        marker=marker,
        sort_keys=sort_keys,
        sort_dirs=sort_dirs,
        fields=fields,
        **kwargs
    )


def create_workbook(values):
    return IMPL.create_workbook(values)


def update_workbook(name, values):
    return IMPL.update_workbook(name, values)


def create_or_update_workbook(name, values):
    return IMPL.create_or_update_workbook(name, values)


def delete_workbook(name, namespace=None):
    IMPL.delete_workbook(name, namespace)


def delete_workbooks(**kwargs):
    IMPL.delete_workbooks(**kwargs)


# Workflow definitions.

def get_workflow_definition(identifier, namespace='', fields=()):
    return IMPL.get_workflow_definition(
        identifier,
        namespace=namespace,
        fields=fields
    )


def get_workflow_definition_by_id(id, fields=()):
    return IMPL.get_workflow_definition_by_id(id, fields=fields)


def load_workflow_definition(name, namespace='', fields=()):
    """Unlike get_workflow_definition this method is allowed to return None."""
    return IMPL.load_workflow_definition(name, namespace, fields=fields)


def get_workflow_definitions(limit=None, marker=None, sort_keys=None,
                             sort_dirs=None, fields=None, **kwargs):
    return IMPL.get_workflow_definitions(
        limit=limit,
        marker=marker,
        sort_keys=sort_keys,
        sort_dirs=sort_dirs,
        fields=fields,
        **kwargs
    )


def create_workflow_definition(values):
    return IMPL.create_workflow_definition(values)


def update_workflow_definition(identifier, values):
    return IMPL.update_workflow_definition(identifier, values)


def create_or_update_workflow_definition(name, values):
    return IMPL.create_or_update_workflow_definition(name, values)


def delete_workflow_definition(identifier, namespace=''):
    IMPL.delete_workflow_definition(identifier, namespace)


def delete_workflow_definitions(**kwargs):
    IMPL.delete_workflow_definitions(**kwargs)


# Dynamic actions.

def get_dynamic_action_definition(identifier, namespace='', fields=()):
    return IMPL.get_dynamic_action_definition(identifier, fields, namespace)


def load_dynamic_action_definition(identifier, namespace='', fields=()):
    return IMPL.load_dynamic_action_definition(identifier, fields, namespace)


def create_dynamic_action_definition(values):
    return IMPL.create_dynamic_action_definition(values)


def update_dynamic_action_definition(identifier, values, namespace=''):
    return IMPL.update_dynamic_action_definition(identifier, values, namespace)


def get_dynamic_action_definitions(limit=None, marker=None, sort_keys=None,
                                   sort_dirs=None, fields=None, **kwargs):
    return IMPL.get_dynamic_action_definitions(
        limit=limit,
        marker=marker,
        sort_keys=sort_keys,
        sort_dirs=sort_dirs,
        fields=fields,
        **kwargs
    )


def delete_dynamic_action_definition(identifier, namespace=''):
    return IMPL.delete_dynamic_action_definition(identifier, namespace)


def delete_dynamic_action_definitions(**kwargs):
    return IMPL.delete_dynamic_action_definitions(**kwargs)


# Code sources.

def get_code_source(identifier, namespace='', fields=()):
    return IMPL.get_code_source(identifier, fields, namespace=namespace)


def load_code_source(identifier, namespace='', fields=()):
    return IMPL.load_code_source(identifier, fields, namespace=namespace)


def create_code_source(values):
    return IMPL.create_code_source(values)


def update_code_source(identifier, values, namespace=''):
    return IMPL.update_code_source(identifier, values, namespace=namespace)


def get_code_sources(limit=None, marker=None, sort_keys=None,
                     sort_dirs=None, fields=None, **kwargs):
    return IMPL.get_code_sources(
        limit=limit,
        marker=marker,
        sort_keys=sort_keys,
        sort_dirs=sort_dirs,
        fields=fields,
        **kwargs
    )


def delete_code_source(identifier, namespace=''):
    return IMPL.delete_code_source(identifier, namespace=namespace)


def delete_code_sources(**kwargs):
    return IMPL.delete_code_sources(**kwargs)


# Action definitions.

def get_action_definition_by_id(id, fields=()):
    return IMPL.get_action_definition_by_id(id, fields=fields)


def get_action_definition(name, fields=(), namespace=''):
    return IMPL.get_action_definition(name, fields=fields, namespace=namespace)


def load_action_definition(name, fields=(), namespace=''):
    """Unlike get_action_definition this method is allowed to return None."""
    action_def = IMPL.load_action_definition(
        name,
        fields=fields,
        namespace=namespace
    )

    # If action definition was not found in the workflow namespace,
    # check in the default namespace
    if action_def:
        return action_def

    return IMPL.load_action_definition(name, fields=fields, namespace='')


def get_action_definitions(limit=None, marker=None, sort_keys=None,
                           sort_dirs=None, **kwargs):
    return IMPL.get_action_definitions(
        limit=limit,
        marker=marker,
        sort_keys=sort_keys,
        sort_dirs=sort_dirs,
        **kwargs
    )


def create_action_definition(values):
    return IMPL.create_action_definition(values)


def update_action_definition(identifier, values):
    return IMPL.update_action_definition(identifier, values)


def create_or_update_action_definition(name, values):
    return IMPL.create_or_update_action_definition(name, values)


def delete_action_definition(name, namespace=''):
    return IMPL.delete_action_definition(name, namespace=namespace)


def delete_action_definitions(**kwargs):
    return IMPL.delete_action_definitions(**kwargs)


# Action executions.

def get_action_execution(id, fields=(), insecure=False):
    return IMPL.get_action_execution(id, fields=fields, insecure=insecure)


def load_action_execution(name, fields=()):
    """Unlike get_action_execution this method is allowed to return None."""
    return IMPL.load_action_execution(name, fields=fields)


def get_action_executions(**kwargs):
    return IMPL.get_action_executions(**kwargs)


def create_action_execution(values):
    return IMPL.create_action_execution(values)


def update_action_execution(id, values, insecure=False):
    return IMPL.update_action_execution(id, values, insecure)


def create_or_update_action_execution(id, values):
    return IMPL.create_or_update_action_execution(id, values)


def update_action_execution_heartbeat(id):
    return IMPL.update_action_execution_heartbeat(id)


def delete_action_execution(id):
    return IMPL.delete_action_execution(id)


def delete_action_executions(**kwargs):
    IMPL.delete_action_executions(**kwargs)


# Workflow executions.

def get_workflow_execution(id, fields=()):
    return IMPL.get_workflow_execution(id, fields=fields)


def load_workflow_execution(name, fields=()):
    """Unlike get_workflow_execution this method is allowed to return None."""
    return IMPL.load_workflow_execution(name, fields=fields)


def get_workflow_executions(limit=None, marker=None, sort_keys=None,
                            sort_dirs=None, **kwargs):
    return IMPL.get_workflow_executions(
        limit=limit,
        marker=marker,
        sort_keys=sort_keys,
        sort_dirs=sort_dirs,
        **kwargs
    )


def create_workflow_execution(values):
    return IMPL.create_workflow_execution(values)


def update_workflow_execution(id, values):
    return IMPL.update_workflow_execution(id, values)


def create_or_update_workflow_execution(id, values):
    return IMPL.create_or_update_workflow_execution(id, values)


def delete_workflow_execution(id):
    return IMPL.delete_workflow_execution(id)


def delete_workflow_executions(**kwargs):
    IMPL.delete_workflow_executions(**kwargs)


def update_workflow_execution_state(**kwargs):
    return IMPL.update_workflow_execution_state(**kwargs)


# Tasks executions.

def get_task_execution(id, fields=()):
    return IMPL.get_task_execution(id, fields=fields)


def load_task_execution(id, fields=()):
    """Unlike get_task_execution this method is allowed to return None."""
    return IMPL.load_task_execution(id, fields=fields)


def get_task_executions(limit=None, marker=None, sort_keys=None,
                        sort_dirs=None, **kwargs):
    return IMPL.get_task_executions(
        limit=limit,
        marker=marker,
        sort_keys=sort_keys,
        sort_dirs=sort_dirs,
        **kwargs
    )


def get_task_executions_count(**kwargs):
    return IMPL.get_task_executions_count(**kwargs)


def get_completed_task_executions(**kwargs):
    return IMPL.get_completed_task_executions(**kwargs)


def get_completed_task_executions_as_batches(**kwargs):
    return IMPL.get_completed_task_executions_as_batches(**kwargs)


def get_incomplete_task_executions(**kwargs):
    return IMPL.get_incomplete_task_executions(**kwargs)


def get_incomplete_task_executions_count(**kwargs):
    return IMPL.get_incomplete_task_executions_count(**kwargs)


def create_task_execution(values):
    return IMPL.create_task_execution(values)


def update_task_execution(id, values):
    return IMPL.update_task_execution(id, values)


def create_or_update_task_execution(id, values):
    return IMPL.create_or_update_task_execution(id, values)


def delete_task_execution(id):
    return IMPL.delete_task_execution(id)


def delete_task_executions(**kwargs):
    return IMPL.delete_task_executions(**kwargs)


def update_task_execution_state(**kwargs):
    return IMPL.update_task_execution_state(**kwargs)


# Delayed calls.

def get_delayed_calls_to_start(time, batch_size=None):
    return IMPL.get_delayed_calls_to_start(time, batch_size)


def get_overdue_calls(time):
    return IMPL.get_overdue_calls(time)


def create_delayed_call(values):
    return IMPL.create_delayed_call(values)


def delete_delayed_call(id):
    return IMPL.delete_delayed_call(id)


def update_delayed_call(id, values, query_filter=None):
    return IMPL.update_delayed_call(id, values, query_filter)


def get_delayed_call(id):
    return IMPL.get_delayed_call(id)


def get_delayed_calls(**kwargs):
    return IMPL.get_delayed_calls(**kwargs)


def get_delayed_calls_count(**kwargs):
    return IMPL.get_delayed_calls_count(**kwargs)


def delete_delayed_calls(**kwargs):
    return IMPL.delete_delayed_calls(**kwargs)


# Scheduled jobs.

def get_scheduled_jobs_to_start(time, batch_size=None):
    return IMPL.get_scheduled_jobs_to_start(time, batch_size)


def create_scheduled_job(values):
    return IMPL.create_scheduled_job(values)


def delete_scheduled_job(id):
    return IMPL.delete_scheduled_job(id)


def update_scheduled_job(id, values, query_filter=None):
    return IMPL.update_scheduled_job(id, values, query_filter)


def get_scheduled_job(id):
    return IMPL.get_scheduled_job(id)


def get_scheduled_jobs(**kwargs):
    return IMPL.get_scheduled_jobs(**kwargs)


def delete_scheduled_jobs(**kwargs):
    return IMPL.delete_scheduled_jobs(**kwargs)


def get_scheduled_jobs_count(**kwargs):
    return IMPL.get_scheduled_jobs_count(**kwargs)


# Cron triggers.

def get_cron_trigger(identifier, fields=()):
    return IMPL.get_cron_trigger(identifier, fields=fields)


def get_cron_trigger_by_id(id, fields=()):
    return IMPL.get_cron_trigger_by_id(id, fields=fields)


def load_cron_trigger(identifier, fields=()):
    """Unlike get_cron_trigger this method is allowed to return None."""
    return IMPL.load_cron_trigger(identifier, fields=fields)


def get_cron_triggers(**kwargs):
    return IMPL.get_cron_triggers(**kwargs)


def get_next_cron_triggers(time):
    return IMPL.get_next_cron_triggers(time)


def get_expired_executions(expiration_time, limit=None, columns=()):
    return IMPL.get_expired_executions(
        expiration_time,
        limit,
        columns
    )


def get_running_expired_sync_action_executions(expiration_time,
                                               limit, session=None):
    return IMPL.get_running_expired_sync_action_executions(
        expiration_time,
        limit
    )


def get_superfluous_executions(max_finished_executions, limit=None,
                               columns=()):
    return IMPL.get_superfluous_executions(
        max_finished_executions,
        limit,
        columns
    )


def create_cron_trigger(values):
    return IMPL.create_cron_trigger(values)


def update_cron_trigger(identifier, values, query_filter=None):
    return IMPL.update_cron_trigger(identifier, values,
                                    query_filter=query_filter)


def create_or_update_cron_trigger(identifier, values):
    return IMPL.create_or_update_cron_trigger(identifier, values)


def delete_cron_trigger(identifier):
    return IMPL.delete_cron_trigger(identifier)


def delete_cron_triggers(**kwargs):
    return IMPL.delete_cron_triggers(**kwargs)


# Environments.

def get_environment(name, fields=()):
    return IMPL.get_environment(name, fields=fields)


def load_environment(name, fields=()):
    """Unlike get_environment this method is allowed to return None."""
    return IMPL.load_environment(name, fields=fields)


def get_environments(limit=None, marker=None, sort_keys=None,
                     sort_dirs=None, **kwargs):
    return IMPL.get_environments(
        limit=limit,
        marker=marker,
        sort_keys=sort_keys,
        sort_dirs=sort_dirs,
        **kwargs
    )


def create_environment(values):
    return IMPL.create_environment(values)


def update_environment(name, values):
    return IMPL.update_environment(name, values)


def create_or_update_environment(name, values):
    return IMPL.create_or_update_environment(name, values)


def delete_environment(name):
    IMPL.delete_environment(name)


def delete_environments(**kwargs):
    IMPL.delete_environments(**kwargs)


# Resource members.


def create_resource_member(values):
    return IMPL.create_resource_member(values)


def get_resource_member(resource_id, res_type, member_id):
    return IMPL.get_resource_member(resource_id, res_type, member_id)


def get_resource_members(resource_id, res_type):
    return IMPL.get_resource_members(resource_id, res_type)


def update_resource_member(resource_id, res_type, member_id, values):
    return IMPL.update_resource_member(
        resource_id,
        res_type,
        member_id,
        values
    )


def delete_resource_member(resource_id, res_type, member_id):
    IMPL.delete_resource_member(resource_id, res_type, member_id)


def delete_resource_members(**kwargs):
    IMPL.delete_resource_members(**kwargs)


# Event triggers.

def get_event_trigger(id, insecure=False, fields=()):
    return IMPL.get_event_trigger(id, insecure, fields=fields)


def load_event_trigger(id, insecure=False, fields=()):
    return IMPL.load_event_trigger(id, insecure, fields=fields)


def get_event_triggers(insecure=False, limit=None, marker=None, sort_keys=None,
                       sort_dirs=None, fields=None, **kwargs):
    return IMPL.get_event_triggers(
        insecure=insecure,
        limit=limit,
        marker=marker,
        sort_keys=sort_keys,
        sort_dirs=sort_dirs,
        fields=fields,
        **kwargs
    )


def create_event_trigger(values):
    return IMPL.create_event_trigger(values)


def update_event_trigger(id, values):
    return IMPL.update_event_trigger(id, values)


def delete_event_trigger(id):
    return IMPL.delete_event_trigger(id)


def delete_event_triggers(**kwargs):
    return IMPL.delete_event_triggers(**kwargs)


# Locks.

def create_named_lock(name):
    return IMPL.create_named_lock(name)


def get_named_locks(limit=None, marker=None):
    return IMPL.get_named_locks(limit=limit, marker=marker)


def delete_named_lock(lock_id):
    return IMPL.delete_named_lock(lock_id)


@contextlib.contextmanager
def named_lock(name):
    with IMPL.named_lock(name):
        yield


def get_maintenance_status():
    return IMPL.get_maintenance_status()


def update_maintenance_status(status):
    return IMPL.update_maintenance_status(status)



# Copyright 2013 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
# Copyright 2016 - Brocade Communications Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistral import exceptions as exc
from mistral import expressions as expr
from mistral.tests.unit import base

DATA = {
    "server": {
        "id": "03ea824a-aa24-4105-9131-66c48ae54acf",
        "name": "cloud-fedora",
        "status": "ACTIVE"
    },
    "status": "OK"
}

SERVERS = {
    "servers": [
        {'name': 'centos'},
        {'name': 'ubuntu'},
        {'name': 'fedora'}
    ]
}


class ExpressionsTest(base.BaseTest):
    def test_evaluate_complex_expressions(self):
        data = {
            'a': 1,
            'b': 2,
            'c': 3,
            'd': True,
            'e': False,
            'f': 10.1,
            'g': 10,
            'h': [1, 2, 3, 4, 5],
            'i': 'We are OpenStack!',
            'j': 'World',
            'k': 'Mistral',
            'l': 'awesome',
            'm': 'the way we roll'
        }

        test_cases = [
            ('<% $.a + $.b * $.c %>', 7),
            ('<%($.a + $.b) * $.c %>', 9),
            ('<% $.d and $.e %>', False),
            ('<% $.f > $.g %>', True),
            ('<% $.h.len() >= 5 %>', True),
            ('<% $.h.len() >= $.b + $.c %>', True),
            ('<% 100 in $.h %>', False),
            ('<% $.a in $.h%>', True),
            ('<% ''OpenStack'' in $.i %>', True),
            ('Hello, <% $.j %>!', 'Hello, World!'),
            ('<% $.k %> is <% $.l %>!', 'Mistral is awesome!'),
            ('This is <% $.m %>.', 'This is the way we roll.'),
            ('<% 1 + 1 = 3 %>', False)
        ]

        for expression, expected in test_cases:
            actual = expr.evaluate_recursively(expression, data)

            self.assertEqual(expected, actual)

    def test_evaluate_recursively(self):
        task_spec_dict = {
            'parameters': {
                'p1': 'My string',
                'p2': '<% $.param2 %>',
                'p3': ''
            },
            'publish': {
                'new_key11': 'new_key1'
            }
        }

        modified_task = expr.evaluate_recursively(
            task_spec_dict,
            {'param2': 'val32'}
        )

        self.assertDictEqual(
            {
                'parameters': {
                    'p1': 'My string',
                    'p2': 'val32',
                    'p3': ''
                },
                'publish': {
                    'new_key11': 'new_key1'
                }
            },
            modified_task
        )

    def test_evaluate_recursively_arbitrary_dict(self):
        context = {
            "auth_token": "123",
            "project_id": "mistral"
        }

        data = {
            "parameters": {
                "parameter1": {
                    "name1": "<% $.auth_token %>",
                    "name2": "val_name2"
                },
                "param2": [
                    "var1",
                    "var2",
                    "/servers/<% $.project_id %>/bla"
                ]
            },
            "token": "<% $.auth_token %>"
        }

        applied = expr.evaluate_recursively(data, context)

        self.assertDictEqual(
            {
                "parameters": {
                    "parameter1": {
                        "name1": "123",
                        "name2": "val_name2"
                    },
                    "param2": ["var1", "var2", "/servers/mistral/bla"]
                },
                "token": "123"
            },
            applied
        )

    def test_evaluate_recursively_environment(self):
        environment = {
            'host': 'vm1234.example.com',
            'db': 'test',
            'timeout': 600,
            'verbose': True,
            '__actions': {
                'std.sql': {
                    'conn': 'mysql://admin:secret@<% env().host %>'
                            '/<% env().db %>'
                }
            }
        }

        context = {
            '__env': environment
        }

        defaults = context['__env']['__actions']['std.sql']
        applied = expr.evaluate_recursively(defaults, context)
        expected = 'mysql://admin:secret@vm1234.example.com/test'

        self.assertEqual(expected, applied['conn'])

    def test_validate_jinja_with_yaql_context(self):
        self.assertRaises(exc.JinjaGrammarException,
                          expr.validate,
                          '{{ $ }}')

    def test_validate_mixing_jinja_and_yaql(self):
        self.assertRaises(exc.ExpressionGrammarException,
                          expr.validate,
                          '<% $.a %>{{ _.a }}')

        self.assertRaises(exc.ExpressionGrammarException,
                          expr.validate,
                          '{{ _.a }}<% $.a %>')

    def test_evaluate_mixing_jinja_and_yaql(self):
        actual = expr.evaluate('<% $.a %>{{ _.a }}', {'a': 'b'})

        self.assertEqual('<% $.a %>b', actual)

        actual = expr.evaluate('{{ _.a }}<% $.a %>', {'a': 'b'})

        self.assertEqual('b<% $.a %>', actual)



# Copyright 2020 - Nokia Software.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from mistral.db.v2 import api as db_api
from mistral.tests.unit.api import base

FILE_CONTENT = "test file"

UPDATED_FILE_CONTENT = "updated content"

MODULE_NAME = 'modulename%s'
NAMESPACE = "NS"


class TestCodeSourcesController(base.APITest):
    def _create_code_source(self, module_name, file_content,
                            namespace=NAMESPACE, expect_errors=False):
        return self.app.post(
            '/v2/code_sources?name=%s&namespace=%s' % (module_name, namespace),
            file_content,
            headers={'Content-Type': 'text/plain'},
            expect_errors=expect_errors
        )

    def _delete_code_source(self, id, namespace=NAMESPACE):
        return self.app.delete(
            '/v2/code_sources/%s?namespace=%s' % (id, namespace)
        )

    def setUp(self):
        super(TestCodeSourcesController, self).setUp()

        self.addCleanup(db_api.delete_code_sources)

    def test_post(self):
        mod_name = MODULE_NAME % 'create'

        resp = self._create_code_source(mod_name, FILE_CONTENT)

        self.assertEqual(201, resp.status_int)

        code_src = resp.json

        self.assertEqual(mod_name, code_src['name'])
        self.assertEqual(FILE_CONTENT, code_src['content'])
        self.assertEqual(1, code_src['version'])
        self.assertEqual(NAMESPACE, code_src['namespace'])

    def test_put(self):
        mod_name = MODULE_NAME % 'update'

        self._create_code_source(mod_name, FILE_CONTENT)

        resp = self.app.put(
            '/v2/code_sources?identifier=%s&namespace=%s' %
            (mod_name, NAMESPACE),
            UPDATED_FILE_CONTENT,
            headers={'Content-Type': 'text/plain'}
        )

        self.assertEqual(200, resp.status_int)

        code_src = resp.json

        self.assertEqual(mod_name, code_src['name'])
        self.assertEqual(UPDATED_FILE_CONTENT, code_src['content'])
        self.assertEqual(2, code_src['version'])
        self.assertEqual(NAMESPACE, code_src['namespace'])

    def test_delete(self):
        mod_name = MODULE_NAME % 'delete'

        resp = self._create_code_source(mod_name, FILE_CONTENT)

        self.assertEqual(201, resp.status_int)
        self.assertEqual(mod_name, resp.json['name'])

        # Make sure the object is in DB.
        self.assertIsNotNone(
            db_api.load_code_source(mod_name, namespace=NAMESPACE)
        )

        self._delete_code_source(mod_name)

        # Make sure the object was deleted from DB.
        self.assertIsNone(
            db_api.load_code_source(mod_name, namespace=NAMESPACE)
        )

    def test_post_duplicate(self):
        mod_name = MODULE_NAME % 'duplicate'

        self._create_code_source(mod_name, FILE_CONTENT)

        resp = self._create_code_source(
            mod_name,
            FILE_CONTENT, expect_errors=True
        )

        self.assertEqual(409, resp.status_int)
        self.assertIn('Duplicate entry for CodeSource', resp)

        self._delete_code_source(mod_name)

    def test_get(self):
        mod_name = MODULE_NAME % 'get'

        self._create_code_source(mod_name, FILE_CONTENT)

        resp = self.app.get(
            '/v2/code_sources/%s?namespace=%s' % (mod_name, NAMESPACE)
        )

        resp_json = resp.json

        self.assertEqual(200, resp.status_int)

        self.assertEqual(mod_name, resp_json['name'])
        self.assertEqual(FILE_CONTENT, resp_json['content'])
        self.assertEqual(1, resp_json['version'])
        self.assertEqual(NAMESPACE, resp_json['namespace'])

    def test_get_all(self):
        mod_name = MODULE_NAME % 'getall'
        mod2_name = MODULE_NAME % '2getall'

        self._create_code_source(mod_name, FILE_CONTENT)
        self._create_code_source(mod2_name, FILE_CONTENT)

        resp = self.app.get(
            '/v2/code_sources',
            params='namespace=%s' % NAMESPACE
        )

        resp_json = resp.json

        self.assertEqual(200, resp.status_int)

        code_sources = resp_json['code_sources']

        self.assertEqual(2, len(code_sources))



# Copyright 2019 - Nokia Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
"""

import sys


def _print_help():
    print("\nUsage: <script_name> <input_file_name> <output_file_name>\n")
    print(
        'The script processes a Mistral profiler log file (<input_file_name)\n'
        'and generates a report into a file (<output_file_name>) that\n'
        'contains statistics about each profiler trace: \n'
        '-------------------------------------------------------------\n'
        ' Total time | Max time | Avg time | Occurrences | Trace name \n'
        '-------------------------------------------------------------\n'
        ' ...          ...        ...        ...           ...\n'
    )


def main():
    try:
        in_file_name = str(sys.argv[1])
        out_file_name = str(sys.argv[2])
    except:
        _print_help()

        return "Failed to parse arguments."

    print('Ranking profiled methods...')

    in_f = open(in_file_name, 'r')
    out_f = open(out_file_name, 'w')

    # {trace_name: [total_time, max_time, occurrences]}
    d = dict()

    with in_f:
        for line in in_f:
            tokens = line.split()

            # Skip all "-start" lines that don't contain a duration in seconds.
            # Processing only "-stop" lines.
            if len(tokens[1]) > 10:
                continue

            trace_name = tokens[5]
            trace_name = trace_name[0:len(trace_name) - 5]

            duration = float(tokens[1])

            if trace_name not in d:
                d[trace_name] = [duration, duration, 1]
            else:
                l = d[trace_name]

                l[0] = l[0] + duration
                l[2] = l[2] + 1

                if duration > l[1]:
                    l[1] = duration

    result = sorted(d.items(), key=lambda x: x[1][0], reverse=True)

    out_f.write('Total time | Max time | Avg time | Occurrences | Trace name\n')
    out_f.write('-' * 90)
    out_f.write('-\n')

    for item in result:
        out_f.write(
            '{0:<12.3f} {1:<10.3f} {2:<10.3f} {3:<13d} {4}\n'.format(
                item[1][0],
                item[1][1],
                item[1][0] / item[1][2],
                item[1][2],
                item[0]
            )
        )

    out_f.close()

    print("Ranking file was successfully created: %s" % out_file_name)


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistral import expressions


NONEMPTY_STRING = {
    "type": "string",
    "minLength": 1
}

UNIQUE_STRING_LIST = {
    "type": "array",
    "items": NONEMPTY_STRING,
    "uniqueItems": True,
    "minItems": 1
}

POSITIVE_INTEGER = {
    "type": "integer",
    "minimum": 0
}

POSITIVE_NUMBER = {
    "type": "number",
    "minimum": 0.0
}

EXPRESSION = {
    "oneOf": [{
        "type": "string",
        "pattern": "^%s\\s*$" % expressions.patterns[name]
    } for name in expressions.patterns]
}

EXPRESSION_CONDITION = {
    "type": "object",
    "minProperties": 1,
    "patternProperties": {
        r"^\w+$": EXPRESSION
    }
}

ANY = {
    "anyOf": [
        {"type": "array"},
        {"type": "boolean"},
        {"type": "integer"},
        {"type": "number"},
        {"type": "object"},
        {"type": "string"}
    ]
}

ANY_NULLABLE = {
    "anyOf": [
        {"type": "null"},
        {"type": "array"},
        {"type": "boolean"},
        {"type": "integer"},
        {"type": "number"},
        {"type": "object"},
        {"type": "string"}
    ]
}

NONEMPTY_DICT = {
    "type": "object",
    "minProperties": 1,
    "patternProperties": {
        r"^\w+$": ANY_NULLABLE
    }
}

ONE_KEY_DICT = {
    "type": "object",
    "minProperties": 1,
    "maxProperties": 1,
    "patternProperties": {
        r"^\w+$": ANY_NULLABLE
    }
}

STRING_OR_EXPRESSION_CONDITION = {
    "oneOf": [
        NONEMPTY_STRING,
        EXPRESSION_CONDITION
    ]
}

EXPRESSION_OR_POSITIVE_INTEGER = {
    "oneOf": [
        EXPRESSION,
        POSITIVE_INTEGER
    ]
}

EXPRESSION_OR_BOOLEAN = {
    "oneOf": [
        EXPRESSION,
        {"type": "boolean"}
    ]
}


UNIQUE_STRING_OR_EXPRESSION_CONDITION_LIST = {
    "type": "array",
    "items": STRING_OR_EXPRESSION_CONDITION,
    "uniqueItems": True,
    "minItems": 1
}

VERSION = {
    "anyOf": [
        NONEMPTY_STRING,
        POSITIVE_INTEGER,
        POSITIVE_NUMBER
    ]
}

WORKFLOW_TYPE = {
    "enum": ["reverse", "direct"]
}

STRING_OR_ONE_KEY_DICT = {
    "oneOf": [
        NONEMPTY_STRING,
        ONE_KEY_DICT
    ]
}

UNIQUE_STRING_OR_ONE_KEY_DICT_LIST = {
    "type": "array",
    "items": STRING_OR_ONE_KEY_DICT,
    "uniqueItems": True,
    "minItems": 1
}



# Copyright 2022 - NetCracker Technology Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from mistral.db.v2 import api as db_api
from mistral.services import workflows as wf_service

from mistral.tests.unit.engine import base
from mistral.workflow import states


class TaskSkipTest(base.EngineTestCase):

    def test_basic_task_skip(self):
        workflow = """
        version: '2.0'
        wf:
          tasks:
            t1:
              action: std.fail
              on-skip: t2
              on-success: t3
            t2:
              action: std.noop
            t3:
              action: std.noop
        """
        wf_service.create_workflows(workflow)

        wf_ex = self.engine.start_workflow('wf')
        self.await_workflow_error(wf_ex.id)

        # Check that on-skip branch was not executed
        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)
            task_execs = wf_ex.task_executions

        self.assertEqual(1, len(task_execs))
        t1_ex = self._assert_single_item(
            task_execs,
            name='t1',
            state=states.ERROR
        )

        # Skip t1 and wait for wf to complete
        self.engine.rerun_workflow(t1_ex.id, skip=True)
        self.await_workflow_success(wf_ex.id)

        # Check that on-skip branch was executed
        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)
            task_execs = wf_ex.task_executions

        self.assertEqual(2, len(task_execs))
        self._assert_single_item(task_execs, name='t1', state=states.SKIPPED)
        self._assert_single_item(task_execs, name='t2', state=states.SUCCESS)

    def test_task_skip_on_workflow_tail(self):
        workflow = """
        version: '2.0'
        wf:
          tasks:
            t0:
              action: std.noop
              on-success: t1
            t1:
              action: std.fail
        """
        wf_service.create_workflows(workflow)

        wf_ex = self.engine.start_workflow('wf')
        self.await_workflow_error(wf_ex.id)

        # Check that on-skip branch was not executed
        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)
            task_execs = wf_ex.task_executions

        self.assertEqual(2, len(task_execs))
        t1_ex = self._assert_single_item(
            task_execs,
            name='t1',
            state=states.ERROR
        )

        # Skip t1 and wait for wf to complete
        self.engine.rerun_workflow(t1_ex.id, skip=True)
        self.await_workflow_success(wf_ex.id)

        # Check that on-skip branch was executed
        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)
            task_execs = wf_ex.task_executions

        self.assertEqual(2, len(task_execs))
        self._assert_single_item(task_execs, name='t0', state=states.SUCCESS)
        self._assert_single_item(task_execs, name='t1', state=states.SKIPPED)

    def test_skip_subworkflow(self):
        workflow = """
        version: '2.0'
        wf:
          tasks:
            t0:
              action: std.noop
              on-success: t1
            t1:
              workflow: subwf
        subwf:
          tasks:
            t0:
              action: std.fail
        """
        wf_service.create_workflows(workflow)

        wf_ex = self.engine.start_workflow('wf')
        self.await_workflow_error(wf_ex.id)

        # Check that on-skip branch was not executed
        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)
            task_execs = wf_ex.task_executions

        self.assertEqual(2, len(task_execs))
        t1_ex = self._assert_single_item(
            task_execs,
            name='t1',
            state=states.ERROR
        )

        # Skip t1 and wait for wf to complete
        self.engine.rerun_workflow(t1_ex.id, skip=True)
        self.await_workflow_success(wf_ex.id)

        # Check that on-skip branch was executed
        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)
            task_execs = wf_ex.task_executions

        self.assertEqual(2, len(task_execs))
        self._assert_single_item(task_execs, name='t0', state=states.SUCCESS)
        self._assert_single_item(task_execs, name='t1', state=states.SKIPPED)

    def test_publish_on_skip(self):
        workflow = """
        version: '2.0'
        wf:
          tasks:
            t0:
              action: std.noop
              on-success: t1
            t1:
              action: std.fail
              publish:
                success: 1
              publish-on-error:
                error: 1
              publish-on-skip:
                skip: 1
        """
        wf_service.create_workflows(workflow)

        wf_ex = self.engine.start_workflow('wf')
        self.await_workflow_error(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)
            task_execs = wf_ex.task_executions

            self.assertEqual(states.ERROR, wf_ex.state)
            self.assertEqual(2, len(task_execs))

            t1_ex = self._assert_single_item(
                task_execs,
                name='t1',
                state=states.ERROR
            )

            publish_before_skip = {"error": 1}
            self.assertDictEqual(publish_before_skip, t1_ex.published)

        # Skip t1 and wait for wf to complete
        self.engine.rerun_workflow(t1_ex.id, skip=True)
        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)
            task_execs = wf_ex.task_executions

            t1_ex = self._assert_single_item(
                task_execs,
                name='t1',
                state=states.SKIPPED
            )

            publish_after_skip = {"skip": 1}
            self.assertDictEqual(publish_after_skip, t1_ex.published)
            self.assertDictEqual(publish_after_skip, wf_ex.output)

    def test_retry_dont_conflict_with_skip(self):
        workflow = """
        version: '2.0'
        wf:
          tasks:
            t1:
              action: std.fail
              on-skip: skip
              retry:
                count: 2
                delay: 0
            skip:
              action: std.noop
        """
        wf_service.create_workflows(workflow)

        wf_ex = self.engine.start_workflow('wf')
        self.await_workflow_error(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)
            task_execs = wf_ex.task_executions

        self.assertEqual(1, len(task_execs))

        t1_ex = self._assert_single_item(
            task_execs,
            name='t1',
            state=states.ERROR
        )

        self.engine.rerun_workflow(t1_ex.id, skip=True)
        self.await_workflow_success(wf_ex.id)



#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import pecan
from pecan import rest
import wsmeext.pecan as wsme_pecan

from mistral.api.controllers import resource
from mistral import context
from mistral.db.v2 import api as db_api
from mistral.services import maintenance
from mistral.utils import rest_utils


class Maintenance(resource.Resource):
    """Maintenance resource."""

    status = str

    @classmethod
    def sample(cls):
        return cls(
            status="1234"
        )


class MaintenanceController(rest.RestController):

    @pecan.expose('json')
    def get(self):
        context.set_ctx(None)

        maintenance_status = db_api.get_maintenance_status()

        return {
            'status': maintenance_status
        }

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(
        Maintenance,
        body=Maintenance,
        status_code=200
    )
    def put(self, new_maintenance_status):
        context.set_ctx(None)

        maintenance.change_maintenance_mode(new_maintenance_status.status)

        return new_maintenance_status
